from .build import EVALUATOR_REGISTRY, build_evaluator
from .ct_evaluation import CTEvaluator
from .evaluator import DatasetEvaluator, inference_on_dataset
from .metrics import *
from .qdess_evaluation import QDESSEvaluator
from .sem_seg_evaluation import SemSegEvaluator

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import logging
from typing import Union

import numpy as np
import tensorflow as tf

from medsegpy.loss.utils import get_activation, reduce_tensor
from medsegpy.utils import env

try:
    import tf.keras.backend as K
except ImportError:
    import keras.backend as K

logger = logging.getLogger(__name__)


class DiceLoss:
    def __init__(
        self,
        weights=None,
        activation: str = None,
        flatten: Union[str, bool] = False,
        remove_background: bool = False,
        reduction: str = "mean",
        reduction_axis=None,
        eps: float = None,
        name: str = None,
    ):
        """
        Args:
            weights (array-like, optional): Class weighting to use.
            activation (bool, optional): Activation function to apply.
                One of `"sigmoid"` | `"softmax"` | `"none"`. If "none",
                no activation will be applied to input.
            flatten (`bool` or `str`, optional): Dimensions to collapse
                into the spatial dimension prior to dice computation.
                One of `True` | `"batch"` | `"channel"`.
                If `True`, flatten both batch and channel dimensions.
            remove_background (bool, optional): If `True`, drop the first channel
                from both y_true and y_pred tensors.
            reduction (str, optional): Reduction to apply to the loss.
            One of `"none"` | `"mean"` | `"sum"`.
                * `"none"`: no reduction applied
                * `"mean"`: The weighted mean of the output is taken
                * `"sum"`: The output will be summed
            reduction_axis (int, optional): The axis to apply reduction to.
            eps (float, optional): Smoothing factor. Defaults to `keras.backend.epsilon()`.
            name (str, optional): Optional name for the op.
        """
        self.__name__ = name if name else type(self).__name__
        self.weights = np.asarray(weights) if weights is not None else weights
        self.remove_background = remove_background
        self.eps = K.epsilon() if eps is None else eps

        if activation not in ("none", None):
            self.activation, self._act_fn, act_args = get_activation(activation)
        else:
            self.activation, self._act_fn, act_args = "none", None, ()
        self._act_kwargs = {}
        if "axis" in act_args:
            # Activation computed over channel dimension
            self._act_kwargs["axis"] = -1

        assert isinstance(flatten, bool) or flatten in ("batch", "channel")
        if weights is not None and flatten == "channel":
            raise ValueError("Cannot flatten across 'channel' dimension and weight channels")
        self.flatten = flatten

        assert reduction in (None, "none", "mean", "sum")
        self.reduction = reduction
        self.reduction_axis = reduction_axis

    def __call__(self, y_true, y_pred):
        if self._act_fn is not None:
            y_pred = self._act_fn(y_pred, **self._act_kwargs)

        if self.remove_background:
            y_true, y_pred = y_true[..., 1:], y_pred[..., 1:]
        if env.is_tf2():
            y_true = tf.dtypes.cast(y_true, y_pred.dtype)

        spatial_dims = tuple(range(1, K.ndim(y_pred) - 1))  # spatial dims
        if self.flatten == "batch":
            dims = (0,) + spatial_dims
        elif self.flatten == "channel":
            dims = spatial_dims + (K.ndim(y_pred) - 1,)
        elif self.flatten:
            dims = (0,) + spatial_dims + (K.ndim(y_pred) - 1,)
        else:
            dims = spatial_dims

        intersection = K.sum(y_true * y_pred, axis=dims)
        overlap = K.sum(y_true, axis=dims) + K.sum(y_pred, axis=dims)
        dice = (2.0 * intersection + self.eps) / (overlap + self.eps)
        loss = 1 - dice  # largest shape: (N, C)

        weights = K.variable(self.weights) if self.weights is not None else None
        return reduce_tensor(
            loss, reduction=self.reduction, axis=self.reduction_axis, weights=weights
        )



import logging

import numpy as np
from keras.initializers import he_normal
from keras.layers import BatchNormalization as BN
from keras.layers import Concatenate, Conv2D, Conv2DTranspose, Dropout, Input, MaxPooling2D
from keras.models import Model
from keras.utils import plot_model
from keras.utils.layer_utils import get_source_inputs

logger = logging.getLogger(__name__)


def anisotropic_unet_2d(
    input_size=None,
    input_tensor=None,
    output_mode=None,
    num_filters=None,
    depth=6,
    kernel_size=(3, 3),
    pooling_size=None,
    pooling_ratio=None,
):
    """Generate Unet 2D model compatible with Keras 2

    :param input_size: tuple of input size - format: (height, width, 1)

    :rtype: Keras model

    :raise ValueError if input_size is not tuple or dimensions of input_size do not match
        (height, width, 1)
    """
    from medsegpy import glob_constants

    logger.info("Initializing unet with seed: %s" % str(glob_constants.SEED))
    SEED = glob_constants.SEED
    if input_tensor is None and (type(input_size) is not tuple or len(input_size) != 3):
        raise ValueError("input_size must be a tuple of size (height, width, 1)")

    if num_filters is None:
        nfeatures = [2**feat * 32 for feat in np.arange(depth)]
    else:
        nfeatures = num_filters
        assert len(nfeatures) == depth

    conv_ptr = []

    # input layer
    inputs = input_tensor if input_tensor is not None else Input(input_size)

    # step down convolutional layers
    unpooling_sizes = []
    pool = inputs
    for depth_cnt in range(depth):

        conv = Conv2D(
            nfeatures[depth_cnt],
            kernel_size,
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(pool)
        conv = Conv2D(
            nfeatures[depth_cnt],
            kernel_size,
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.0)(conv)

        conv_ptr.append(conv)

        # Only maxpool till penultimate depth
        if depth_cnt < depth - 1:
            x_shape = conv.shape.as_list()
            pool_size = (
                pooling_size if pooling_size else __get_pooling_size__(x_shape, pooling_ratio)
            )
            unpooling_sizes.append(pool_size)
            pool = MaxPooling2D(pool_size=pool_size, padding="same")(conv)

    # step up convolutional layers
    for depth_cnt in range(depth - 2, -1, -1):
        deconv_shape = conv_ptr[depth_cnt].shape.as_list()
        deconv_shape[0] = None
        unpooling_size = unpooling_sizes[depth_cnt]
        up = Concatenate(axis=3)(
            [
                Conv2DTranspose(
                    nfeatures[depth_cnt],
                    kernel_size,
                    padding="same",
                    strides=unpooling_size,
                    kernel_initializer=he_normal(seed=SEED),
                )(conv),
                conv_ptr[depth_cnt],
            ]
        )

        conv = Conv2D(
            nfeatures[depth_cnt],
            kernel_size,
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(up)
        conv = Conv2D(
            nfeatures[depth_cnt],
            kernel_size,
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.00)(conv)

    # combine features
    # this if statement is required for legacy purposes
    # some weights were trained with a joint activation,
    # which makes it difficult to load weights effectively
    if output_mode is not None:
        recon = Conv2D(
            1,
            (1, 1),
            padding="same",
            activation=output_mode,
            kernel_initializer=he_normal(seed=SEED),
        )(conv)
    else:
        recon = conv

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)

    model = Model(inputs=inputs, outputs=[recon])

    return model


def __get_pooling_size__(x_shape: list, pool_ratio=None):
    yres = x_shape[1]
    xres = x_shape[2]

    pooling_size = []
    for r in yres, xres:
        if r % 2 == 0:
            pooling_size.append(2)
        else:
            pooling_size.append(3)

    pooling_size = tuple(pooling_size)

    if not pool_ratio:
        return pooling_size

    if yres > xres:
        return pooling_size[0] * pool_ratio, pooling_size[1]
    else:
        return pooling_size[0], pooling_size[1] * pool_ratio


if __name__ == "__main__":
    save_path = "../imgs/anisotropic_unet2d.png"
    m = anisotropic_unet_2d(input_size=(72, 288, 1), kernel_size=(3, 3))
    plot_model(m, to_file=save_path, show_shapes=True)



import os
import random
import re
import sys
import typing
import warnings
from time import localtime, strftime
from typing import Union

from medsegpy.utils import io_utils

sys.path.append("../")

sys.path.append("../")

# Change based on dataset used for analysis. Do not use any digits
DATASET_NAME = "oai_imorphics"

# =======DO NOT CHANGE CONSTANTS BELOW=======
K_REGEX_PATTERN = "_cv-k[0-9]+"
K_BIN_FILENAME_BASE = DATASET_NAME + "_cv-k%d"
K_BIN_SAVE_DIRECTORY = os.path.dirname(os.path.realpath(__file__))


class CrossValidationProcessor:
    """A processor class for handling computation for cross-validation experiments"""

    __k_NUM_VALID_BINS_KEY = "num_valid_bins"
    __k_NUM_TEST_BINS_KEY = "num_test_bins"

    def __init__(self, k_or_filepath: Union[int, str], **kwargs):
        """
        :param k_or_filepath: an int describing the number of bins (k)
                              or a filepath to an existing cross-validation file
        :param \**kwargs: See below

        :keyword num_valid_bins: Number of bins allocated for validation. Must be used with keyword num_test_bins.
        :keyword num_test_bins: Number of bins allocated for testing. Must be used with keyword num_valid_bins.
        """
        if type(k_or_filepath) not in [int, str]:
            raise ValueError(
                "`k_or_filepath` must be either the k value (int) or the filepath (str)"
            )

        if type(k_or_filepath) is int:
            k = k_or_filepath
            filepaths = self.__get_cross_validation_files(k)

            if not filepaths:
                raise ValueError("No file found for k=%d. Create file with `create_cv_bins.py`" % k)

            if len(filepaths) != 1:
                raise ValueError(
                    "Multiple files corresponding to k=%d. Please specify explicit filepath."
                )

            filepath = filepaths[0]
        else:
            filepath = k_or_filepath

        if not os.path.isfile(filepath):
            raise FileNotFoundError("File %s not found" % filepath)

        k_from_filename = self.__get_k_from_file(filepath)
        cv_data = io_utils.load_pik(filepath)
        assert len(cv_data) == k_from_filename, "Corrupted file: mismatch bins"

        self._filepath = filepath
        self._k = k_from_filename
        self._bin_files = cv_data

        # These fields should be appropriately populated by instance methods
        self.num_valid_bins = 0
        self.num_test_bins = 0
        self.bins_split = None

        # Handle kwargs
        if self.__k_NUM_VALID_BINS_KEY in kwargs and self.__k_NUM_TEST_BINS_KEY in kwargs:
            self.init_cv_experiments(kwargs.get("num_valid_bins"), kwargs.get("num_test_bins"))

        if (self.__k_NUM_VALID_BINS_KEY not in kwargs and self.__k_NUM_TEST_BINS_KEY in kwargs) or (
            self.__k_NUM_VALID_BINS_KEY in kwargs and self.__k_NUM_TEST_BINS_KEY not in kwargs
        ):
            raise ValueError(
                "%s and %s must be specified together"
                % (self.__k_NUM_VALID_BINS_KEY, self.__k_NUM_TEST_BINS_KEY)
            )

    def init_cv_experiments(self, num_valid_bins=1, num_test_bins=1):
        """
        Organize k bins into training, validation, testing bins for k/num_test_bins experiments

        :param num_valid_bins: Number of bins for validation
        :param num_test_bins: Number of bins for testing. Must be perfect divisor of k (k % num_test_bins = 0)

        """
        k = self.k

        # number of holdout bins cannot exceed k
        num_holdout = num_valid_bins + num_test_bins
        if num_holdout > k:
            raise ValueError("Number of holdout bins (validation + test) must be < k")

        # inference sets cannot overlap in cross-validation
        if k % num_test_bins != 0:
            raise ValueError(
                "There can be no overlap in test bins across different cross-validation trials"
            )

        if num_holdout / k > 0.4:
            warnings.warn(
                "%0.1f holdout - validation: %0.1f, test: %0.1f"
                % (num_holdout / k * 100, num_valid_bins / k * 100, num_test_bins / k * 100)
            )
        num_train = k - num_holdout

        test_bin_start_ind = 0
        exps_bin_division = []
        for i in range(int(k / num_test_bins)):
            valid_bin_start_ind = test_bin_start_ind + num_test_bins
            train_bin_start_ind = valid_bin_start_ind + num_valid_bins

            test_bins = [ind % k for ind in range(test_bin_start_ind, valid_bin_start_ind)]
            valid_bins = [ind % k for ind in range(valid_bin_start_ind, train_bin_start_ind)]
            train_bins = [
                ind % k for ind in range(train_bin_start_ind, train_bin_start_ind + num_train)
            ]

            assert (
                len(set(train_bins) & set(valid_bins)) == 0
            ), "Training and Validation bins must be mutually exclusive"
            assert (
                len(set(train_bins) & set(test_bins)) == 0
            ), "Training and Test bins must be mutually exclusive"
            assert (
                len(set(valid_bins) & set(test_bins)) == 0
            ), "Validation and Test bins must be mutually exclusive"

            exps_bin_division.append((train_bins, valid_bins, test_bins))

            test_bin_start_ind += num_test_bins

        # Check to make sure all test bins are mutually exclusive
        temp = []
        for d in exps_bin_division:
            temp.append(d[-1])

        for i in range(len(temp)):
            for j in range(i + 1, len(temp)):
                assert (
                    len(set(temp[i]) & set(temp[j])) == 0
                ), "Test bins %d and %d not mutually exclusive - %d overlap" % (
                    i,
                    j,
                    len(set(temp[i]) & set(temp[j])),
                )

        self.num_valid_bins = num_valid_bins
        self.num_test_bins = num_test_bins
        self.bins_split = exps_bin_division

    def run(self):
        """
        Yields training, validation, testing files and train, validation, testing bin indexes as single tuple
        """
        bins_split = self.bins_split
        bins_files = self.bin_files

        if not bins_split:
            raise RuntimeError(
                "Cross-validation experiment is not initialized. Call init_cv_experiments"
            )

        for bin_inds in bins_split:
            assert len(bin_inds) == 3, "Expected 3 sets of bin indices - train, valid, test"
            train_bins, valid_bins, test_bins = tuple(bin_inds)
            train_files, valid_files, test_files = self.get_fnames(
                (train_bins, valid_bins, test_bins)
            )

            yield train_files, valid_files, test_files, train_bins, valid_bins, test_bins

    def get_fnames(self, bin_inds):
        """
        Get filepaths for training, validation, and testing
        :param bin_inds: A tuple of indexes for training, validation, and testing respectively

        :return: A tuple of lists. Each list consists of filepaths for training, validation, testing respectively
        """
        bin_files = self.bin_files
        train_inds, valid_inds, test_inds = bin_inds

        train_files = [bin_files[x] for x in train_inds]
        train_files = [filepath for x in train_files for filepath in x]

        valid_files = [bin_files[x] for x in valid_inds]
        valid_files = [filepath for x in valid_files for filepath in x]

        test_files = [bin_files[x] for x in test_inds]
        test_files = [filepath for x in test_files for filepath in x]

        return train_files, valid_files, test_files

    @staticmethod
    def __get_cross_validation_files(k: int) -> tuple:
        """
        Get list of cross-validation files for given k value
        :param k: The k value (k-fold cross validation)

        :return: A tuple of filepaths to .cv files
        """
        save_directory = K_BIN_SAVE_DIRECTORY
        base_name = K_BIN_FILENAME_BASE % k

        files = []
        for f in os.listdir(save_directory):
            if base_name in f:
                files.append(os.path.join(save_directory, f))

        return tuple(files)

    @staticmethod
    def __get_k_from_file(filepath: str) -> int:
        """
        Get k value (k-fold) from filepath
        :param filepath: A filepath to a cross-validation (.cv) file

        :return: An int
        """
        filename = os.path.basename(filepath)

        matches = re.findall(K_REGEX_PATTERN, filename)
        if len(matches) > 1:
            warnings.warn("Multiple matches found - using match at 0th index")
        match = matches[0]

        return int(re.findall("[0-9]+", match)[0])

    @property
    def filepath(self):
        return self._filepath

    @property
    def k(self):
        return self._k

    @property
    def bin_files(self):
        return self._bin_files


class CrossValidationFileGenerator:
    """A generator class for creating file for cross validation data"""

    def __init__(
        self,
        k_bins: int,
        data_paths: typing.Collection[str],
        dataset_tag: str,
        overwrite: bool = False,
    ):
        """
        :param k_bins: Number of bins to create for k-fold cross-validation
        :param data_paths: Collection of paths where data is stored. Invalid filepaths will be ignored
        :param dataset_tag: An identifier for the data used for cross validation
        :param overwrite: Whether existing file should be overwritten (optional)
        """
        self.k_bins = k_bins
        self.data_paths = data_paths

        fname = dataset_tag + "_cv-k%d-%s.cv" % (k_bins, strftime("%Y-%m-%d-%H-%M-%S", localtime()))
        self.save_path = os.path.join(K_BIN_SAVE_DIRECTORY, fname)
        tmp_path = os.path.join(K_BIN_SAVE_DIRECTORY, "%s_cv-k%d.cv" % ("tmp", k_bins))

        save_path = self.save_path

        # Overwrite file if exists
        if os.path.isfile(save_path):
            if not overwrite:
                raise FileExistsError("File %s exists." % save_path)

        bins = self.generate_bins()

        # save data to filepath
        io_utils.save_pik(bins, tmp_path)

        try:
            # Verify list
            self.verify_bins(tmp_path, k_bins)
        except Exception as e:
            os.remove(tmp_path)
            raise e

        os.remove(tmp_path)
        io_utils.save_pik(bins, self.save_path)

    def verify_bins(self, filepath, expected_k):
        cv_processor = CrossValidationProcessor(filepath)

        # verify number of bins are expected length
        assert cv_processor.k == expected_k, "Expected %d bins, got %d bins" % (
            expected_k,
            cv_processor.k,
        )
        k_bins = cv_processor.k
        bins = cv_processor.bin_files

        bin_to_pid_dict = dict()
        bin_to_scanid_dict = dict()

        for bin_id in range(k_bins):
            bin = bins[bin_id]
            pids = []
            scan_ids = []
            for filepath in bin:
                file_info = self.__get_file_info(
                    os.path.basename(filepath), os.path.dirname(filepath)
                )
                pids.append(file_info["pid"])
                scan_ids.append(file_info["scanid"])

            bin_to_pid_dict[bin_id] = list(set(pids))
            bin_to_scanid_dict[bin_id] = list(set(scan_ids))

        max_num_pids = max([len(bin_to_pid_dict[bin_id]) for bin_id in range(k_bins)])
        min_num_pids = min([len(bin_to_pid_dict[bin_id]) for bin_id in range(k_bins)])
        assert (
            max_num_pids - min_num_pids <= 1
        ), "Difference in number of subjects between bins should be <= 1"

        # max_num_scan_ids = max([len(bin_to_scanid_dict[bin_id]) for bin_id in range(k)])
        # min_num_scan_ids = min([len(bin_to_scanid_dict[bin_id]) for bin_id in range(k)])
        # assert max_num_scan_ids - min_num_scan_ids <= 2  # each pid has 2 scans

    def generate_bins(self):
        # get pid dictionary
        pids_dict = self.__parse_pids()
        pids = list(pids_dict.keys())

        # Shuffle pids in random order
        random.shuffle(pids)

        # Allocate each pid to a bin
        bins_list = self.__get_bins_list(len(pids), self.k_bins)
        pid_bin_map = dict()
        for i in range(len(pids)):
            pid_bin_map[pids[i]] = bins_list[i]

        bins = (
            []
        )  # stores filepaths for scans/scan slices from patients with pid corresponding to this bin
        for i in range(self.k_bins):
            bins.append([])
        for dp in self.data_paths:
            for fname in os.listdir(dp):
                if fname.endswith(".im"):
                    im_info = self.__get_file_info(fname, dp)

                    # Get bin that this pid should be in
                    bin_id = pid_bin_map[im_info["pid"]]
                    filepath = os.path.join(dp, im_info["fname"])

                    bins[bin_id].append(filepath)

        # Check that bins are mutually exclusive
        for i in range(len(bins)):
            for j in range(i + 1, len(bins)):
                if len(set(bins[i]) & set(bins[j])) != 0:
                    overlap = list(set(bins[i]) & set(bins[j]))
                    overlap.sort()
                    for fp in overlap:
                        print(fp)
                    print(pid_bin_map)
                    raise ValueError("Bins %d and %d not exclusive" % (i, j))

            # Check for duplicates
        self.check_duplicates(bins)

        return bins

    def __parse_pids(self):
        """Get patient identifiers"""
        pids = dict()
        for dp in self.data_paths:
            for fname in os.listdir(dp):
                if fname.endswith(".im"):
                    im_info = self.__get_file_info(fname, dp)
                    curr_pid = im_info["pid"]
                    if curr_pid in pids.keys():
                        assert dp == pids[curr_pid], "dirpath mismatch. Expected: %s, got %s" % (
                            dp,
                            pids[curr_pid],
                        )
                    else:
                        pids[curr_pid] = dp

        pids_dict = pids.copy()

        return pids_dict

    def __get_file_info(self, fname, dirpath):
        fname, ext = os.path.splitext(fname)
        f_data = fname.split("-")
        scan_id = f_data[0]
        pid_timepoint_split = scan_id.split("_")
        pid = pid_timepoint_split[0]
        f_aug_slice = f_data[1].split("_")
        data = {
            "pid": pid,
            "timepoint": int(pid_timepoint_split[1][1:]),
            "aug": int(f_aug_slice[0][3:]),
            "slice": int(f_aug_slice[1]),
            "fname": fname,
            "impath": os.path.join(dirpath, "%s.%s" % (fname, "im")),
            "segpath": os.path.join(dirpath, "%s.%s" % (fname, "seg")),
            "scanid": scan_id,
        }
        assert data["pid"] == fname[:7], str(data)

        return data

    def __get_bins_list(self, num_pids, k):
        num_bins = [int(num_pids / k)] * k
        remainder_num_elements = num_pids % k

        # randomly pick which bins take on additional elements
        remainder_bin_list = random.sample(range(k), remainder_num_elements)
        for bin_id in remainder_bin_list:
            num_bins[bin_id] += 1

        assert max(num_bins) - min(num_bins) <= 1
        assert len(num_bins) == k

        # Each pid should get a bin
        bin_ids = [[bin_id] * num_bins[bin_id] for bin_id in range(k)]

        # flatten list
        bin_ids = [x for id_list in bin_ids for x in id_list]

        assert len(bin_ids) == num_pids

        return bin_ids

    @staticmethod
    def check_duplicates(x_list: list):
        for i in range(len(x_list)):
            if len(x_list[i]) != len(set(x_list[i])):
                raise ValueError("Duplicates in list %d" % i)


# if __name__ == '__main__':
#     pass
#     # print(get_cv_experiments(6, num_test_bins=2))



import logging
from copy import deepcopy

from keras import callbacks as kc

from medsegpy.utils import env

try:
    import wandb
    import wandb.wandb_run

    _WANDB_AVAILABLE = True
except ImportError:  # pragma: no-cover
    wandb = None
    _WANDB_AVAILABLE = False


__all__ = ["lr_callback", "LossHistory", "WandBLogger"]

logger = logging.getLogger(__name__)


def lr_callback(optimizer):
    """Wrapper for learning rate tensorflow metric.

    Args:
        optimizer: Optimizer used for training.

    Returns:
        func: To be wrapped in metric or callback.
    """

    def lr(y_true, y_pred):
        return optimizer.lr

    return lr


class LossHistory(kc.Callback):
    """A Keras callback to log training history"""

    def on_train_begin(self, logs=None):
        self.val_losses = []
        self.losses = []
        self.epoch = []

    def on_epoch_end(self, epoch, logs=None):
        self.val_losses.append(logs.get("val_loss", float("nan")))
        self.losses.append(logs.get("loss"))
        self.epoch.append(epoch + 1)

        metrics = " - ".join(
            [
                "{}: {:0.4f}".format(k, v) if v >= 1e-3 else "{}: {:0.4e}".format(k, v)
                for k, v in logs.items()
            ]
        )
        logger.info("Epoch {} - {}".format(epoch + 1, metrics))


class WandBLogger(kc.Callback):
    """A Keras callback to log to weights and biases.

    Currently only supports logging scalars.
    """

    def __init__(self, period: int = 20, experiment="auto", **kwargs):
        """
        Args:
            period (int, optional): Logging period.
            experiment (`wandb.wandb_run.Run` | `str` | `None`): The experiment run.
                If ``"auto"``, a run will only be created if ``wandb.run`` is None.
                If ``None``, a run will be created.
            **kwargs: Options to pass to ``wandb.init()`` to create run. Ignored
                if ``experiment`` specified.
        """
        if not env.supports_wandb():
            raise ValueError(
                "Weights & Biases is not supported. "
                "Install package via `pip install wandb`. "
                "See documentation https://docs.wandb.com/ "
            )
        assert isinstance(experiment, wandb.wandb_run.Run) or experiment in ("auto", None)
        if (not wandb.run and experiment == "auto") or experiment is None:
            wandb.init(**kwargs)

        assert isinstance(period, int) and period > 0, "`period` must be int >0"

        self._period = period

    def on_train_begin(self, logs=None):
        self._step = 0

    def on_batch_end(self, batch_idx, logs=None):
        self._step += 1
        if not logs or self._step % self._period != 0:
            return

        wandb.log(logs, step=self._step)

    def on_epoch_end(self, epoch, logs=None):
        logs = deepcopy(logs)
        logs["epoch"] = epoch
        wandb.log(logs, step=self._step)



"""Model loading utils.

Adapted from `keras.saving.model_config`.
https://github.com/tensorflow/tensorflow/blob/e5bf8de410005de06a7ff5393fafdf832ef1d4ad/tensorflow/python/keras/saving/model_config.py#L55
"""
import json

try:
    import yaml  # noqa: E402
except ImportError:
    yaml = None


def model_from_config(config, custom_objects=None):
    """Instantiates a Keras model from its config.
    Arguments:
      config: Configuration dictionary.
      custom_objects: Optional dictionary mapping names
          (strings) to custom classes or functions to be
          considered during deserialization.
    Returns:
      A Keras model instance (uncompiled).
    Raises:
      TypeError: if `config` is not a dictionary.
    """
    if isinstance(config, list):
        raise TypeError(
            "`model_from_config` expects a dictionary, not a list. "
            "Maybe you meant to use "
            "`Sequential.from_config(config)`?"
        )
    from .layers import deserialize  # noqa

    return deserialize(config, custom_objects=custom_objects)


def model_from_yaml(yaml_string, custom_objects=None):
    """Parses a yaml model configuration file and returns a model instance.
    Arguments:
      yaml_string: YAML string encoding a model configuration.
      custom_objects: Optional dictionary mapping names
          (strings) to custom classes or functions to be
          considered during deserialization.
    Returns:
      A Keras model instance (uncompiled).
    Raises:
      ImportError: if yaml module is not found.
    """
    if yaml is None:
        raise ImportError("Requires yaml module installed (`pip install pyyaml`).")
    config = yaml.load(yaml_string)
    from .layers import deserialize  # noqa

    return deserialize(config, custom_objects=custom_objects)


def model_from_json(json_string, custom_objects=None):
    """Parses a JSON model configuration file and returns a model instance.
    Arguments:
      json_string: JSON string encoding a model configuration.
      custom_objects: Optional dictionary mapping names
          (strings) to custom classes or functions to be
          considered during deserialization.
    Returns:
      A Keras model instance (uncompiled).
    """
    config = json.loads(json_string)
    from .layers import deserialize  # noqa

    return deserialize(config, custom_objects=custom_objects)



""" Attention Layers

The following layers implement an attention gating module based on the
paper "Attention U-Net: Learning Where to Look For the Pancreas"
(Oktay et al.). The code below is based on a PyTorch implementation of
this technique by the paper's authors:

https://github.com/ozan-oktay/Attention-Gated-Networks/tree/a96edb72622274f6705097d70cfaa7f2bf818a5a

Each layer has 2D and 3D versions. Only the 2D versions have been tested
so far.
"""
from typing import Dict, Sequence, Union

import numpy as np
from keras import backend as K
from keras.layers import BatchNormalization as BN
from keras.layers import Concatenate, Conv2D, Conv3D, Layer, Multiply, UpSampling2D, UpSampling3D


class _CreateGatingSignalNDim(Layer):
    def __init__(
        self,
        dimension: int,
        out_channels: int,
        kernel_size: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict],
        activation: str,
        add_batchnorm: bool,
        **kwargs
    ):
        """
        This layer creates the first gating signal for attention based on
        a feature map. This feature map should contain contextual information
        that will help the network focus on important regions in the image.
        In the paper, the feature map chosen for a U-Net is the coarsest
        feature map at end of the encoding arch.

        The layer performs a simple operation to transform the feature map
        into the first gating signal:

        Convolution --> Activation --> BatchNorm

        Args:
            dimension: the dimension of the model's input images
            out_channels: the number of channels for the gating signal
            kernel_size: the kernel size used in the convolutional layer
            kernel_initializer: the method for initializing the weights of the
                                    convolutional layer
            activation: the activation function used after the convolutional
                            layer
            add_batchnorm: specifies if batch normalization should be used
        """
        super(_CreateGatingSignalNDim, self).__init__(**kwargs)

        # Store parameters
        self.dimension = dimension
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.kernel_initializer = kernel_initializer
        self.activation = activation
        self.add_batchnorm = add_batchnorm

        if self.dimension == 2:
            conv_type = Conv2D
        elif self.dimension == 3:
            conv_type = Conv3D
        else:
            raise ValueError("Only 2D and 3D are supported")

        if isinstance(self.kernel_size, tuple) or isinstance(self.kernel_size, list):
            assert len(self.kernel_size) == self.dimension, (
                "If list/tuple, kernel_size must have length %d" % self.dimension
            )

        self.conv = conv_type(
            self.out_channels,
            self.kernel_size,
            padding="same",
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
        )
        if self.add_batchnorm:
            self.bn = BN(axis=-1, momentum=0.95, epsilon=0.001)

    def build(self, input_shape):
        self.conv.build(input_shape)
        self._trainable_weights = self.conv.trainable_weights
        conv_output_shape = self.conv.compute_output_shape(input_shape)

        if self.add_batchnorm:
            self.bn.build(conv_output_shape)
            self._trainable_weights += self.bn.trainable_weights
        super(_CreateGatingSignalNDim, self).build(input_shape)

    def call(self, inputs):
        outputs = self.conv(inputs)
        if self.add_batchnorm:
            outputs = self.bn(outputs)
        return outputs

    def compute_output_shape(self, input_shape):
        conv_output_shape = self.conv.compute_output_shape(input_shape)
        final_output_shape = conv_output_shape
        if self.add_batchnorm:
            bn_output_shape = self.bn.compute_output_shape(conv_output_shape)
            final_output_shape = bn_output_shape
        return final_output_shape

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update(
            {
                "dimension": self.dimension,
                "out_channels": self.out_channels,
                "kernel_size": self.kernel_size,
                "kernel_initializer": self.kernel_initializer,
                "activation": self.activation,
                "add_batchnorm": self.add_batchnorm,
            }
        )
        return base_cfg


class CreateGatingSignal2D(_CreateGatingSignalNDim):
    def __init__(
        self,
        out_channels: int,
        kernel_size: Union[int, Sequence[int]] = 1,
        kernel_initializer: Union[str, Dict] = "he_normal",
        activation: str = "relu",
        add_batchnorm: bool = True,
        **kwargs
    ):
        super(CreateGatingSignal2D, self).__init__(
            dimension=2,
            out_channels=out_channels,
            kernel_size=kernel_size,
            kernel_initializer=kernel_initializer,
            activation=activation,
            add_batchnorm=add_batchnorm,
            **kwargs
        )


class CreateGatingSignal3D(_CreateGatingSignalNDim):
    def __init__(
        self,
        out_channels: int,
        kernel_size: Union[int, Sequence[int]] = 1,
        kernel_initializer: Union[str, Dict] = "he_normal",
        activation: str = "relu",
        add_batchnorm: bool = True,
        **kwargs
    ):
        super(CreateGatingSignal3D, self).__init__(
            dimension=3,
            out_channels=out_channels,
            kernel_size=kernel_size,
            kernel_initializer=kernel_initializer,
            activation=activation,
            add_batchnorm=add_batchnorm,
            **kwargs
        )


class _GridAttentionModuleND(Layer):
    def __init__(
        self,
        dimension: int,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict],
        **kwargs
    ):
        """
        This layer implements the additive attention gate proposed
        in the paper and displayed in Figure 2 of the paper. The gate
        takes in a feature map and a gating signal, calculates
        the attention coefficients (ranging from 0 to 1), and multiplies
        the feature map and the attention coefficients to down weight
        unimportant feature vectors, based on contextual information
        from the gating signal. The pruned feature map is then
        linearly transformed using a convolutional layer, which is
        followed by a batch normalization layer.

        The formulas for computing the attention coefficients are
        Equations 1 and 2 in the paper. The code below uses the variables
        "theta_x" and "theta_gating" to represent W_x and W_g (and b_g)
        found in Equations 1 and 2. Sigma_1 in Equation 1 is fixed
        as the ReLU activation function and Sigma_2 in Equation 2 is
        fixed as the sigmoid activation function.

        Args:
            dimension: the dimension of the model's input images
            in_channels: the number of channels in the input feature map
            intermediate_channels: F_int (in Figure 2 of the paper)
            sub_sample_factor: the factor by which the input feature map
                                should be downsampled. This should be chosen
                                such that the input feature map is downsampled
                                to the resolution of the gating signal,
                                as described in the paper under the section
                                "Attention Gates in U-Net Model".
            kernel_initializer: the method used for initializing the weights
                                    of convolutional layers
        """
        super(_GridAttentionModuleND, self).__init__(**kwargs)

        # Store parameters
        self.dimension = dimension
        self.in_channels = in_channels
        self.intermediate_channels = intermediate_channels
        self.sub_sample_factor = sub_sample_factor
        self.kernel_initializer = kernel_initializer

        if self.dimension == 2:
            self.conv_type = Conv2D
            self.upsample_type = UpSampling2D
        elif self.dimension == 3:
            self.conv_type = Conv3D
            # NOTE: The authors of the paper state they use trilinear
            # interpolation for upsampling 3D images. However, there
            # does not exist a trilinear interpolation
            # mode for UpSampling3D.
            self.upsample_type = UpSampling3D
        else:
            raise ValueError("Only 2D and 3D are supported")

        if isinstance(self.sub_sample_factor, tuple) or isinstance(self.sub_sample_factor, list):
            assert len(self.sub_sample_factor) == self.dimension, (
                "If list/tuple, sub_sample_factor must have length %d" % self.dimension
            )

        self.theta_x = self.conv_type(
            self.intermediate_channels,
            kernel_size=self.sub_sample_factor,
            strides=self.sub_sample_factor,
            use_bias=False,
            kernel_initializer=self.kernel_initializer,
        )

        self.theta_gating = self.conv_type(
            self.intermediate_channels, kernel_size=1, kernel_initializer=self.kernel_initializer
        )

        self.psi = self.conv_type(1, kernel_size=1, kernel_initializer=self.kernel_initializer)

        self.output_conv = self.conv_type(
            self.in_channels, kernel_size=1, kernel_initializer=self.kernel_initializer
        )

        self.output_bn = BN(axis=-1, momentum=0.95, epsilon=0.001)

    def build(self, input_shape):
        x_shape, gating_signal_shape = input_shape

        # Build theta_x
        self.theta_x.build(x_shape)
        self._trainable_weights = self.theta_x.trainable_weights
        theta_x_output_shape = self.theta_x.compute_output_shape(x_shape)

        # Build theta_gating
        self.theta_gating.build(gating_signal_shape)
        self._trainable_weights += self.theta_gating.trainable_weights
        theta_gating_output_shape = self.theta_gating.compute_output_shape(gating_signal_shape)

        # Build upsample_gating
        up_ratio_gating = np.floor_divide(
            theta_x_output_shape[1:-1], theta_gating_output_shape[1:-1]
        )
        self.upsample_gating = self.upsample_type(size=tuple(up_ratio_gating))
        self.upsample_gating.build(theta_gating_output_shape)
        self._trainable_weights += self.upsample_gating.trainable_weights
        up_gating_output_shape = self.upsample_gating.compute_output_shape(
            theta_gating_output_shape
        )
        assert (
            up_gating_output_shape[1:-1] == theta_x_output_shape[1:-1]
        ), "Cannot upsample output of theta_gating to match size of output of theta_x"

        # Build psi
        self.psi.build(theta_x_output_shape)
        self._trainable_weights += self.psi.trainable_weights
        psi_output_shape = self.psi.compute_output_shape(theta_x_output_shape)

        # Build upsample_attn_coeff
        up_ratio_attn_coeff = np.floor_divide(x_shape[1:-1], psi_output_shape[1:-1])
        self.upsample_attn_coeff = self.upsample_type(size=tuple(up_ratio_attn_coeff))
        self.upsample_attn_coeff.build(psi_output_shape)
        self._trainable_weights += self.upsample_attn_coeff.trainable_weights
        up_coeff_output_shape = self.upsample_attn_coeff.compute_output_shape(psi_output_shape)
        assert (
            up_coeff_output_shape[1:-1] == x_shape[1:-1]
        ), "Cannot upsample output of psi to match size of input feature map (x)"

        # Build output_conv
        self.output_conv.build(x_shape)
        self._trainable_weights += self.output_conv.trainable_weights
        output_conv_output_shape = self.output_conv.compute_output_shape(x_shape)

        # Build output_bn
        self.output_bn.build(output_conv_output_shape)
        self._trainable_weights += self.output_bn.trainable_weights

        super(_GridAttentionModuleND, self).build(input_shape)

    def call(self, inputs):
        x, gating_signal = inputs
        theta_x_out = self.theta_x(x)
        theta_gating_out = self.theta_gating(gating_signal)

        # If theta_gating_out is smaller than theta_x_out,
        # then upsample using UpSampling2D or UpSampling3D. This should
        # not be needed if the right value is chosen for
        # sub_sample_factor.
        up_sampled_gating = self.upsample_gating(theta_gating_out)
        psi_out = self.psi(K.relu(theta_x_out + up_sampled_gating))
        sigmoid_psi = K.sigmoid(psi_out)
        x_size = K.int_shape(x)

        # Need to upsample to size of the input feature map (x), such that
        # the attention coefficients can be multiplied with the input
        # feature map
        up_sampled_attn_coeff = self.upsample_attn_coeff(sigmoid_psi)
        attn_weighted_output = Multiply()(
            [K.repeat_elements(up_sampled_attn_coeff, rep=x_size[-1], axis=-1), x]
        )
        output = self.output_conv(attn_weighted_output)
        output = self.output_bn(output)

        return [output, up_sampled_attn_coeff]

    def compute_output_shape(self, input_shape):
        x_shape, gating_signal_shape = input_shape

        theta_gating_output_shape = self.theta_gating.compute_output_shape(gating_signal_shape)
        up_gating_output_shape = self.upsample_gating.compute_output_shape(
            theta_gating_output_shape
        )
        psi_output_shape = self.psi.compute_output_shape(up_gating_output_shape)
        up_attn_output_shape = self.upsample_attn_coeff.compute_output_shape(psi_output_shape)
        output_conv_output_shape = self.output_conv.compute_output_shape(x_shape)
        output_bn_output_shape = self.output_bn.compute_output_shape(output_conv_output_shape)

        return [output_bn_output_shape, up_attn_output_shape]

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update(
            {
                "dimension": self.dimension,
                "in_channels": self.in_channels,
                "intermediate_channels": self.intermediate_channels,
                "sub_sample_factor": self.sub_sample_factor,
                "kernel_initializer": self.kernel_initializer,
            }
        )
        return base_cfg


class GridAttentionModule2D(_GridAttentionModuleND):
    def __init__(
        self,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]] = 2,
        kernel_initializer: Union[str, Dict] = "he_normal",
        **kwargs
    ):
        super(GridAttentionModule2D, self).__init__(
            dimension=2,
            in_channels=in_channels,
            intermediate_channels=intermediate_channels,
            sub_sample_factor=sub_sample_factor,
            kernel_initializer=kernel_initializer,
            **kwargs
        )


class GridAttentionModule3D(_GridAttentionModuleND):
    def __init__(
        self,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]] = 2,
        kernel_initializer: Union[str, Dict] = "he_normal",
        **kwargs
    ):
        super(GridAttentionModule3D, self).__init__(
            dimension=3,
            in_channels=in_channels,
            intermediate_channels=intermediate_channels,
            sub_sample_factor=sub_sample_factor,
            kernel_initializer=kernel_initializer,
            **kwargs
        )


class _MultiAttentionModuleND(Layer):
    def __init__(
        self,
        dimension: int,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict],
        activation: str,
        **kwargs
    ):
        """
        This layer combines the outputs of two attention gates that
        each receive the same inputs. The outputs are concatenated
        along the channel axis and passed through a convolutional layer
        and a batch normalization layer. The operation is as follows:

        Concatenated Gate Outputs --> Convolution --> Activation --> BatchNorm

        Due to the concatenation operation, the layer learns
        multi-dimensional attention coefficients, as described in the
        paper under the section "Attention Gates for Image Analysis".
        After concatenation, the total attention coefficients are the
        coefficients for each gate, concatenated along the channel
        dimension. The "multi-dimensional" aspect comes from how
        each attention coefficient is now a vector with two elements.

        The number of attention gates (2) was chosen because the authors
        of the paper used the same number in their PyTorch implementation
        (GitHub URL included above).

        Args:
            activation: the activation function after the convolutional layer
            all other parameters: same as in _GridAttentionModuleND
        """
        super(_MultiAttentionModuleND, self).__init__(**kwargs)

        # Store parameters
        self.dimension = dimension
        self.in_channels = in_channels
        self.intermediate_channels = intermediate_channels
        self.sub_sample_factor = sub_sample_factor
        self.kernel_initializer = kernel_initializer
        self.activation = activation

        if self.dimension == 2:
            self.conv_type = Conv2D
            self.attn_module_type = GridAttentionModule2D
        elif self.dimension == 3:
            self.conv_type = Conv3D
            self.attn_module_type = GridAttentionModule3D
        else:
            raise ValueError("Only 2D and 3D are supported")

        self.attn_gate_1 = self.attn_module_type(
            in_channels=self.in_channels,
            intermediate_channels=self.intermediate_channels,
            sub_sample_factor=self.sub_sample_factor,
            kernel_initializer=self.kernel_initializer,
        )

        self.attn_gate_2 = self.attn_module_type(
            in_channels=self.in_channels,
            intermediate_channels=self.intermediate_channels,
            sub_sample_factor=self.sub_sample_factor,
            kernel_initializer=self.kernel_initializer,
        )

        self.combine_gates_conv = self.conv_type(
            self.in_channels,
            kernel_size=1,
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
        )
        self.combine_gates_bn = BN(axis=-1, momentum=0.95, epsilon=0.001)

    def build(self, input_shape):
        self.attn_gate_1.build(input_shape)
        self._trainable_weights = self.attn_gate_1.trainable_weights
        self.attn_gate_2.build(input_shape)
        self._trainable_weights += self.attn_gate_2.trainable_weights
        output_attn_shape, _ = self.attn_gate_2.compute_output_shape(input_shape)

        concatenate_gate_shape = list(output_attn_shape)
        concatenate_gate_shape[-1] *= 2
        self.combine_gates_conv.build(concatenate_gate_shape)
        self._trainable_weights += self.combine_gates_conv.trainable_weights
        conv_output_shape = self.combine_gates_conv.compute_output_shape(concatenate_gate_shape)
        self.combine_gates_bn.build(conv_output_shape)
        self._trainable_weights += self.combine_gates_bn.trainable_weights
        super(_MultiAttentionModuleND, self).build(input_shape)

    def call(self, inputs):
        x, gating_signal = inputs
        gate_1_out, attn_coeff_1 = self.attn_gate_1([x, gating_signal])
        gate_2_out, attn_coeff_2 = self.attn_gate_2([x, gating_signal])

        total_gate_outputs = Concatenate(axis=-1)([gate_1_out, gate_2_out])
        total_attn_coeffs = Concatenate(axis=-1)([attn_coeff_1, attn_coeff_2])
        output = self.combine_gates_conv(total_gate_outputs)
        output = self.combine_gates_bn(output)
        return [output, total_attn_coeffs]

    def compute_output_shape(self, input_shape):
        output_attn_shape, coeff_shape = self.attn_gate_2.compute_output_shape(input_shape)
        concatenate_gate_shape = list(output_attn_shape)
        concatenate_coeff_shape = list(coeff_shape)
        concatenate_gate_shape[-1] *= 2
        concatenate_coeff_shape[-1] *= 2
        conv_output_shape = self.combine_gates_conv.compute_output_shape(concatenate_gate_shape)
        bn_output_shape = self.combine_gates_bn.compute_output_shape(conv_output_shape)
        return [bn_output_shape, concatenate_coeff_shape]

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update(
            {
                "dimension": self.dimension,
                "in_channels": self.in_channels,
                "intermediate_channels": self.intermediate_channels,
                "sub_sample_factor": self.sub_sample_factor,
                "kernel_initializer": self.kernel_initializer,
                "activation": self.activation,
            }
        )
        return base_cfg


class MultiAttentionModule2D(_MultiAttentionModuleND):
    def __init__(
        self,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]] = 2,
        kernel_initializer: Union[str, Dict] = "he_normal",
        activation: str = "relu",
        **kwargs
    ):
        super(MultiAttentionModule2D, self).__init__(
            dimension=2,
            in_channels=in_channels,
            intermediate_channels=intermediate_channels,
            sub_sample_factor=sub_sample_factor,
            kernel_initializer=kernel_initializer,
            activation=activation,
            **kwargs
        )


class MultiAttentionModule3D(_MultiAttentionModuleND):
    def __init__(
        self,
        in_channels: int,
        intermediate_channels: int,
        sub_sample_factor: Union[int, Sequence[int]] = 2,
        kernel_initializer: Union[str, Dict] = "he_normal",
        activation: str = "relu",
        **kwargs
    ):
        super(MultiAttentionModule3D, self).__init__(
            dimension=3,
            in_channels=in_channels,
            intermediate_channels=intermediate_channels,
            sub_sample_factor=sub_sample_factor,
            kernel_initializer=kernel_initializer,
            activation=activation,
            **kwargs
        )


class _DeepSupervisionND(Layer):
    def __init__(
        self,
        dimension: int,
        out_channels: int,
        scale_factor: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict],
        **kwargs
    ):
        """
        This layer implements deep-supervision, as mentioned in the paper
        under the section "Attention Gates in U-Net Model". The
        structure of the implementation follows the same structure as
        the author's PyTorch implementation (GitHub URL included above).

        The layer takes a feature map as input, passes it through a
        convolutional layer, and upsamples the output of convolution
        by the user-inputted scale factor.

        In a U-Net, deep-supervision is used to transform the outputs of
        all levels in the decoding arch to have a number of channels equal
        to the number of classes for segmentation, and upsamples these
        outputs to have the same resolution as the input image. These
        transformed outputs are then concatenated along the channel dimension.
        The concatenated outputs are passed through the final step of
        the U-Net to obtain probabilities for segmentation.

        Args:
            dimension: the dimension of the model's input images
            out_channels: the number of channels for the output of the
                            convolutional layer. For a U-Net, this will
                            be the number of classes.
            scale_factor: the factor by which the input feature map is
                            upsampled
            kernel_initializer: the method for initializing the weights of the
                                    convolutional layer
        """
        super(_DeepSupervisionND, self).__init__(**kwargs)

        # Store parameters
        self.dimension = dimension
        self.out_channels = out_channels
        self.scale_factor = scale_factor
        self.kernel_initializer = kernel_initializer

        assert isinstance(self.scale_factor, list) or isinstance(
            self.scale_factor, tuple
        ), "scale_factor must be a list or tuple"

        if isinstance(self.scale_factor, list):
            self.scale_factor = tuple(self.scale_factor)

        if self.dimension == 2:
            self.conv_type = Conv2D
            self.upsample = UpSampling2D(size=self.scale_factor)
        elif self.dimension == 3:
            self.conv_type = Conv3D
            self.upsample = UpSampling3D(size=self.scale_factor)
        else:
            raise ValueError("Only 2D and 3D are supported")

        self.conv = self.conv_type(
            self.out_channels, kernel_size=1, kernel_initializer=self.kernel_initializer
        )

    def build(self, input_shape):
        self.conv.build(input_shape)
        self._trainable_weights = self.conv.trainable_weights
        conv_output_shape = self.conv.compute_output_shape(input_shape)
        self.upsample.build(conv_output_shape)
        self._trainable_weights += self.upsample.trainable_weights
        super(_DeepSupervisionND, self).build(input_shape)

    def call(self, inputs):
        outputs = self.conv(inputs)
        outputs = self.upsample(outputs)
        return outputs

    def compute_output_shape(self, input_shape):
        conv_output_shape = self.conv.compute_output_shape(input_shape)
        upsample_output_shape = self.upsample.compute_output_shape(conv_output_shape)
        return upsample_output_shape

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update(
            {
                "dimension": self.dimension,
                "out_channels": self.out_channels,
                "scale_factor": self.scale_factor,
                "kernel_initializer": self.kernel_initializer,
            }
        )
        return base_cfg


class DeepSupervision2D(_DeepSupervisionND):
    def __init__(
        self,
        out_channels: int,
        scale_factor: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict] = "he_normal",
        **kwargs
    ):
        super(DeepSupervision2D, self).__init__(
            dimension=2,
            out_channels=out_channels,
            scale_factor=scale_factor,
            kernel_initializer=kernel_initializer,
            **kwargs
        )


class DeepSupervision3D(_DeepSupervisionND):
    def __init__(
        self,
        out_channels: int,
        scale_factor: Union[int, Sequence[int]],
        kernel_initializer: Union[str, Dict] = "he_normal",
        **kwargs
    ):
        super(DeepSupervision3D, self).__init__(
            dimension=3,
            out_channels=out_channels,
            scale_factor=scale_factor,
            kernel_initializer=kernel_initializer,
            **kwargs
        )



"""Pooling layers.

MaxPoolingMask2D, MaxPoolingWithArgmax2D, and MaxUnpooling2D are based on are
based on code from the open-source repo below. We thank the authors for making
the code publicly available.

https://github.com/ykamikawa/tf-keras-SegNet/tree/648ee1aa6870e8280a5f24ee193caa585adde9cd.
"""

from keras import backend as K
from keras.layers import Layer
from keras.layers.convolutional import UpSampling2D
from keras.layers.pooling import MaxPooling2D

__all__ = ["MaxPoolingWithArgmax2D", "MaxUnpooling2D", "MaxPooling2D", "UpSampling2D"]


class MaxPoolingWithArgmax2D(Layer):
    def __init__(self, pool_size=(2, 2), strides=(2, 2), padding="same", **kwargs):
        super(MaxPoolingWithArgmax2D, self).__init__(**kwargs)
        self.padding = padding
        self.pool_size = pool_size
        self.strides = strides

    def call(self, inputs, **kwargs):
        padding = self.padding
        pool_size = self.pool_size
        strides = self.strides
        if K.backend() == "tensorflow":
            ksize = [1, pool_size[0], pool_size[1], 1]
            padding = padding.upper()
            strides = [1, strides[0], strides[1], 1]
            output, argmax = K.tf.nn.max_pool_with_argmax(
                inputs, ksize=ksize, strides=strides, padding=padding
            )
        else:
            errmsg = "{} backend is not supported for layer {}".format(
                K.backend(), type(self).__name__
            )
            raise NotImplementedError(errmsg)
        argmax = K.cast(argmax, K.floatx())
        return [output, argmax]

    def compute_output_shape(self, input_shape):
        ratio = (1, self.pool_size[0], self.pool_size[1], 1)
        output_shape = [
            dim // ratio[idx] if dim is not None else None for idx, dim in enumerate(input_shape)
        ]
        output_shape = tuple(output_shape)
        return [output_shape, output_shape]

    def compute_mask(self, inputs, mask=None):
        return self.pool_size[0] * [None]

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update(
            {"padding": self.padding, "pool_size": self.pool_size, "strides": self.strides}
        )
        return base_cfg


class MaxUnpooling2D(Layer):
    def __init__(self, size=(2, 2), **kwargs):
        super(MaxUnpooling2D, self).__init__(**kwargs)
        self.pool_size = size

    def call(self, inputs, output_shape=None):
        updates, mask = inputs[0], inputs[1]
        with K.tf.variable_scope(self.name):
            mask = K.cast(mask, "int32")
            input_shape = K.tf.shape(updates, out_type="int32")
            #  calculation new shape
            if output_shape is None:
                output_shape = (
                    input_shape[0],
                    input_shape[1] * self.pool_size[0],
                    input_shape[2] * self.pool_size[1],
                    input_shape[3],
                )
            self.output_shape1 = output_shape

            # calculation indices for batch, height, width and feature maps
            one_like_mask = K.ones_like(mask, dtype="int32")
            batch_shape = K.concatenate([[input_shape[0]], [1], [1], [1]], axis=0)
            batch_range = K.reshape(K.tf.range(output_shape[0], dtype="int32"), shape=batch_shape)
            b = one_like_mask * batch_range
            y = mask // (output_shape[2] * output_shape[3])
            x = (mask // output_shape[3]) % output_shape[2]
            feature_range = K.tf.range(output_shape[3], dtype="int32")
            f = one_like_mask * feature_range

            # transpose indices & reshape update values to one dimension
            updates_size = K.tf.size(updates)
            indices = K.transpose(K.reshape(K.stack([b, y, x, f]), [4, updates_size]))
            values = K.reshape(updates, [updates_size])
            ret = K.tf.scatter_nd(indices, values, output_shape)
            return ret

    def compute_output_shape(self, input_shape):
        mask_shape = input_shape[1]
        return (
            mask_shape[0],
            mask_shape[1] * self.pool_size[0],
            mask_shape[2] * self.pool_size[1],
            mask_shape[3],
        )

    def get_config(self):
        base_cfg = super().get_config()
        base_cfg.update({"pool_size": self.pool_size})
        return base_cfg



# -*- coding: utf-8 -*-
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# File: transformer.py
import importlib
import inspect
import logging
import pprint
from abc import ABCMeta, abstractmethod
from typing import Sequence, Tuple, Union

import numpy as np
import tqdm

from medsegpy.config import Config
from medsegpy.data.data_utils import generate_poisson_disc_mask

from .transform import (
    CropTransform,
    FillRegionsWithValue,
    MedTransform,
    Swap2DPatches,
    TransformList,
)

__all__ = ["RandomCrop", "TransformGen", "apply_transform_gens", "build_preprocessing"]


# Create custom handler for tqdm
# Copied from "https://stackoverflow.com/questions/38543506/
# change-logging-print-function-to-tqdm-write-so-logging-doesnt-interfere-wit"
class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(TqdmLoggingHandler())


def check_dtype(img: np.ndarray):
    assert isinstance(img, np.ndarray), "[TransformGen] Needs an numpy array, but got a {}!".format(
        type(img)
    )
    assert (
        not isinstance(img.dtype, np.integer) or img.dtype == np.uint8
    ), "[TransformGen] Got image of type {}, " "use uint8 or floating points instead!".format(
        img.dtype
    )
    assert img.ndim > 2, img.ndim


class TransformGen(metaclass=ABCMeta):
    """
    TransformGen takes an array of type float as input.

    It creates a :class:`Transform` based on the given image, sometimes with
    randomness. The transform can then be used to transform images
    or other data (boxes, points, annotations, etc.) associated with it.

    The assumption made in this class is that the image itself is sufficient to
    instantiate a transform. When this assumption is not true, you need to
    create the transforms by your own.

    A list of `TransformGen` can be applied with :func:`apply_transform_gens`.
    """

    def _init(self, params=None):
        if params:
            for k, v in params.items():
                if k != "self" and not k.startswith("_"):
                    setattr(self, k, v)

    @abstractmethod
    def get_transform(self, img):
        pass

    def _rand_range(self, low=1.0, high=None, size=None):
        """
        Uniform float random number between low and high.
        """
        if high is None:
            low, high = 0, low
        if size is None:
            size = []
        return np.random.uniform(low, high, size)

    def __repr__(self):
        """
        Produce something like:
        "MyTransformGen(field1={self.field1}, field2={self.field2})"
        """
        try:
            sig = inspect.signature(self.__init__)
            classname = type(self).__name__
            argstr = []
            for name, param in sig.parameters.items():
                assert (
                    param.kind != param.VAR_POSITIONAL and param.kind != param.VAR_KEYWORD
                ), "The default __repr__ doesn't support *args or **kwargs"
                assert hasattr(self, name), (
                    "Attribute {} not found! "
                    "Default __repr__ only works if attributes match "
                    "the constructor.".format(name)
                )
                attr = getattr(self, name)
                default = param.default
                if default is attr:
                    continue
                argstr.append("{}={}".format(name, pprint.pformat(attr)))
            return "{}({})".format(classname, ", ".join(argstr))
        except AssertionError:
            return super().__repr__()

    __str__ = __repr__


class RandomCrop(TransformGen):
    """Randomly crop a subimage out of an image."""

    def __init__(self, crop_type: str, crop_size: Tuple[float]):
        """
        Args:
            crop_type (str): One of "relative_range", "relative", "absolute".
                See `config/defaults.py` for explanation.
            crop_size (tuple[float]): The relative ratio or absolute pixels of
                width, height, (...)
        """
        super().__init__()
        assert crop_type in ["relative_range", "relative", "absolute"]
        self._init(locals())

    def get_transform(self, img: np.ndarray):
        """
        Args:
            img: a `(...,)H,W` image.
        """
        cdim = len(self.crop_size)
        image_size = img.shape[-cdim:][::-1]
        crop_size = self.get_crop_size(image_size)
        assert all(
            dim >= crop_dim for dim, crop_dim in zip(image_size, crop_size)
        ), "Shape computation in {} has bugs.".format(self)

        # Format: x,y,z,... and w,h,d,...
        image_size = image_size[::-1]
        crop_size = crop_size[::-1]
        coords0 = [
            np.random.randint(img_dim - crop_dim + 1)
            for img_dim, crop_dim in zip(image_size, crop_size)
        ]
        return CropTransform(coords0, crop_size)

    def get_crop_size(self, image_size):
        """
        Args:
            image_size (tuple): width, height, (...)

        Returns:
            crop_size (tuple): width, height, (...) in absolute pixels
        """
        if self.crop_type == "relative":
            crop_size = self.crop_size
            return tuple(int(dim * crop_dim + 0.5) for dim, crop_dim in zip(image_size, crop_size))
        elif self.crop_type == "relative_range":
            crop_size = np.asarray(self.crop_size, dtype=np.float32)
            cdim = len(len(crop_size))
            crop_size = crop_size + np.random.rand(cdim) * (1 - crop_size)
            return tuple(int(dim * crop_dim + 0.5) for dim, crop_dim in zip(image_size, crop_size))
        elif self.crop_type == "absolute":
            return self.crop_size
        else:
            NotImplementedError("Unknown crop type {}".format(self.crop_type))


class CoarseDropout(TransformGen):
    """Randomly select rectangular regions in images and fill them with a
    constant value.

    The rectangular regions are called "holes" in this implementation. All
    rectangular regions will be 2-D for now.

    Ideally, the constant value chosen should be the mean of the image
    intensity distribution, such that this modification does not significantly
    alter the intensity distribution. Significant differences in the input
    intensity distribution between the training set and test set could
    negatively impact the performance of a machine learning model.

    This implementation is adapted from the implementation of coarse dropout
    in:
    https://github.com/albumentations-team/albumentations/blob/master/albumentations/augmentations/transforms.py.
    """

    def __init__(
        self,
        max_holes: int,
        max_height: int,
        max_width: int,
        min_holes: int = None,
        min_height: int = None,
        min_width: int = None,
        img_shape: Sequence[int] = (512, 512),
        max_perc_area_to_remove: float = 0.25,
        fill_value: float = 0.0,
        sampling_pattern: str = "uniform",
        num_precompute: int = 100,
    ):
        """
        Args:
            max_holes: The maximum number of holes to drop out.
            max_height: The maximum height of each hole.
            max_width: The maximum width of each hole.
            min_holes: The minimum number of holes to drop out.
                            If None, the maximum number of holes will
                        be dropped out.
            min_height: The minimum height of each hole.
                            If None, each hole will have maximum height.
            min_width: The minimum width of each hole.
                            If None, each hole will have maximum width.
            img_shape: The height and width of each image.
            max_perc_area_to_remove: The maximum fraction of the image that
                                    will be removed and filled with a constant
                                    value. This is useful to prevent too
                                    much of the image from being modified.
            fill_value: The value used to fill in each hole.
            sampling_pattern: The type of sampling pattern to use when
                                selecting random patches.
                Possible values: "uniform", "poisson"
            num_precompute: The number of masks to precompute.
        """
        super().__init__()
        if min_holes is None:
            min_holes = max_holes
        if min_height is None:
            min_height = max_height
        if min_width is None:
            min_width = max_width
        self._init(locals())

        # Precompute masks
        self.precomputed_masks = []
        if sampling_pattern == "poisson":
            assert min_height == max_height == min_width == max_width, (
                "Only square patches are allowed if sampling pattern is " "'poisson'"
            )

            hole_size = min_width
            # Check max_perc_area_to_remove
            #   - Assuming a hexagonal packing of circles will output the
            #       most number of samples when using Poisson disc sampling.
            #   - From https://mathworld.wolfram.com/CirclePacking.html:
            #       Max packing density when using hexagonal packing is
            #       pi / (2 * sqrt(3))
            max_pos_area = (img_shape[0] * img_shape[1]) * (np.pi / (2 * np.sqrt(3)))
            max_num_patches = max_pos_area // ((np.pi / 2) * (hole_size**2))
            max_pos_perc_area = np.round(
                (max_num_patches * (hole_size**2)) / (img_shape[0] * img_shape[1]), decimals=3
            )
            if max_perc_area_to_remove >= max_pos_perc_area:
                raise ValueError(
                    f"Value of 'max_perc_area_to_remove' "
                    f"(= {max_perc_area_to_remove}) is too large. "
                    f"An overestimate of the maximum possible % area that can "
                    f"be corrupted given {hole_size} x {hole_size} patches "
                    f"is: {max_pos_perc_area}. Make sure "
                    f"'max_perc_area_to_remove' is less than this value."
                )

            # If `sampling_pattern` is "poisson", precompute
            # `num_precompute` masks
            num_samples = ((img_shape[0] * img_shape[1]) * max_perc_area_to_remove) // (
                hole_size**2
            )
            logger.info("Precomputing masks...")
            for _ in tqdm.tqdm(range(num_precompute)):
                _, patch_mask = generate_poisson_disc_mask(
                    (img_shape[0], img_shape[1]),
                    min_distance=hole_size * np.sqrt(2),
                    num_samples=num_samples,
                    patch_size=hole_size,
                    k=10,
                )
                self.precomputed_masks.append(patch_mask)
            logger.info("Finished precomputing masks!")
        else:
            logger.info("Precomputing masks...")
            img_height = img_shape[0]
            img_width = img_shape[1]
            max_area_to_remove = int((img_height * img_width) * max_perc_area_to_remove)
            for _ in tqdm.tqdm(range(num_precompute)):
                patch_mask = np.zeros(img_shape)
                cur_num_holes = 0
                if max_holes < 0:
                    num_holes = np.inf
                else:
                    num_holes = np.random.randint(min_holes, max_holes + 1)
                while (
                    cur_num_holes <= num_holes and np.count_nonzero(patch_mask) < max_area_to_remove
                ):
                    hole_width = np.random.randint(min_width, max_width + 1)
                    hole_height = np.random.randint(min_height, max_height + 1)
                    tl_x = np.random.randint(img_width - hole_width)
                    tl_y = np.random.randint(img_height - hole_height)
                    br_x = tl_x + hole_width
                    br_y = tl_y + hole_height
                    hole_area = (br_x - tl_x) * (br_y - tl_y)
                    if hole_area > max_area_to_remove and np.count_nonzero(patch_mask) == 0:
                        continue
                    patch_mask[tl_y:br_y, tl_x:br_x] = 1
                    cur_num_holes += 1
                self.precomputed_masks.append(patch_mask)
            logger.info("Finished precomputing masks!")

    def get_transform(self, img: np.ndarray) -> MedTransform:
        """
        Determines the holes for 2D images. The location, number, and size
        of these holes are chosen at random, according to the max/min values
        provided by the user.

        The holes will be chosen one by one until either the chosen number of
        holes have been removed or "max_area_to_remove" pixels have been
        removed.

        If multiple 2D images are provided, the holes will be chosen
        independently for each image. Therefore, the chosen holes for one
        2D image may be different from the chosen holes for another 2D image.

        Args:
            img: A N x H x W x C image, containing N images with height H,
                    width W, and consisting of C channels.
        Returns:
            An instance of the MedTransform "FillRegionsWithValue",
                initialized with the appropriate parameters based on the
                input image.
        """
        num_images = img.shape[0]
        hole_mask = np.zeros(img.shape)

        for i in range(num_images):
            # Randomly select one of the precomputed masks
            rand_idx = np.random.randint(self.num_precompute)
            mask = self.precomputed_masks[rand_idx]

            # Randomly rotate by either 0, 90, 180, or 270 degrees
            # counter-clockwise
            mask_rot = np.rot90(mask, k=np.random.randint(4))

            # Get coordinates of patches
            hole_mask[i, :, :, :] = mask_rot[np.newaxis, :, :, np.newaxis]

        return FillRegionsWithValue(hole_mask=hole_mask, fill_value=self.fill_value)


class SwapPatches(TransformGen):
    """Randomly select pairs of non-overlapping patches that will
    be swapped in an iterative manner.

    The size for both patches in a pair will be chosen at random,
    and both patches in a pair will have the same size.

    Furthermore, the two patches in a pair will not overlap.
    """

    def __init__(
        self,
        max_height: int,
        min_height: int = None,
        max_width: int = None,
        min_width: int = None,
        max_iterations: int = 30,
        min_iterations: int = 30,
        is_square: bool = True,
        img_shape: Sequence[int] = (512, 512),
        max_perc_area_to_modify: float = 0.25,
        sampling_pattern: str = "uniform",
        num_precompute: int = 100,
    ):
        """
        Args:
            max_height: The maximum height of each patch. If "is_square" is
                        True, this will also be the maximum width of each
                        patch.
            min_height: The minimum height of each patch. If "is_square" is
                        True, this will also be the minimum width of each
                        patch.
            max_width: The maximum width of each patch. This is only used
                        if "is_square" is False.
            min_width: The minimum width of each patch. This is only used
                        if "is_square" is False.
            max_iterations: The maximum number of pairs of patches to swap. If
                            "max_perc_area_to_modify" is < 1, then the number
                            of iterations may be reduced to ensure only
                            the specified percentage of the area is modified.
            min_iterations: The minimum number of pairs of patches to swap.
            is_square: If True, all patches will be square. Otherwise, all
                        patches will be rectangular, with the possibility of
                        having non-equal heights and widths.
            img_shape: The height and width of each image.
            max_perc_area_to_modify: The maximum fraction of the image that
                                        will be modified. This is useful to
                                        ensure not too much of the image
                                        is changed.
            sampling_pattern: The type of sampling pattern to use when
                                selecting random patches.
                Possible values: "uniform", "poisson"
            num_precompute: The number of masks to precompute.
        """
        super().__init__()
        if sampling_pattern not in ["uniform", "poisson"]:
            raise ValueError(
                f"Invalid value for 'sampling_pattern' ("
                f"Got '{sampling_pattern}'). Must "
                f"be either 'uniform' or 'poisson'."
            )
        if min_height is None:
            min_height = max_height
        if not is_square:
            if max_width is None:
                raise ValueError("Value of 'max_width' must not be None if patch is " "not square")
            if min_width is None:
                min_width = max_width
        self._init(locals())

        self.precomputed_masks = []
        if sampling_pattern == "poisson":
            assert max_height % 2 == 0, f"'max_height' (= {max_height}) must be even"
            assert min_height == max_height, (
                f"If sampling_pattern is 'poisson', min_height (= {min_height}) "
                f"must equal max_height (= {max_height})"
            )
            assert is_square, "Only square patches are allowed if sampling pattern is " "'poisson'"

            patch_size = max_height
            # Check max_perc_area_to_modify
            # -- Assuming a hexagonal packing of circles will output the
            #       most number of samples when using Poisson disc sampling.
            # -- From https://mathworld.wolfram.com/CirclePacking.html:
            #       Max packing density when using hexagonal packing is
            #       pi / (2 * sqrt(3))
            max_pos_area = (img_shape[0] * img_shape[1]) * (np.pi / (2 * np.sqrt(3)))
            max_num_patches = max_pos_area // ((np.pi / 2) * (patch_size**2))
            max_pos_perc_area = np.round(
                (max_num_patches * (patch_size**2)) / (img_shape[0] * img_shape[1]), decimals=3
            )
            if max_perc_area_to_modify >= max_pos_perc_area:
                raise ValueError(
                    f"Value of 'max_perc_area_to_modify' "
                    f"(= {max_perc_area_to_modify}) is too large. "
                    f"An overestimate of the maximum possible % area that can "
                    f"be corrupted given {patch_size} x {patch_size} patches "
                    f"is: {max_pos_perc_area}. Make sure "
                    f"'max_perc_area_to_modify' is less than this value."
                )

            # If `sampling_pattern` is "poisson", precompute
            # `num_precompute` masks
            num_samples = ((img_shape[0] * img_shape[1]) * max_perc_area_to_modify) // (
                patch_size**2
            )
            assert num_samples >= 2, f"Number of samples (= {num_samples}) must be >= 2"
            # Ensure number of samples is even
            if num_samples % 2:
                num_samples -= 1

            logger.info("Precomputing masks...")
            for _ in tqdm.tqdm(range(num_precompute)):
                pd_mask, _ = generate_poisson_disc_mask(
                    (img_shape[0], img_shape[1]),
                    min_distance=patch_size * np.sqrt(2),
                    num_samples=num_samples,
                    patch_size=patch_size,
                    k=10,
                )
                self.precomputed_masks.append(pd_mask)
            logger.info("Finished precomputing masks!")
        else:
            img_height = img_shape[0]
            img_width = img_shape[1]
            area_check = np.zeros((img_height, img_width))
            max_area_to_modify = int((img_height * img_width) * max_perc_area_to_modify)

            logger.info("Precomputing masks...")
            for _ in tqdm.tqdm(range(num_precompute)):
                patch_coords = []
                area_check[:] = 0
                if max_iterations < 0:
                    num_iterations = np.inf
                else:
                    num_iterations = np.random.randint(min_iterations, max_iterations + 1)
                while (len(patch_coords) / 4) <= num_iterations and np.count_nonzero(
                    area_check
                ) < max_area_to_modify:
                    patch_height = np.random.randint(min_height, max_height + 1)
                    if is_square:
                        patch_width = patch_height
                    else:
                        patch_width = np.random.randint(min_width, max_width + 1)
                    # Get coordinates of first patch
                    tl_x_1 = np.random.randint(img_width - patch_width)
                    tl_y_1 = np.random.randint(img_height - patch_height)
                    br_x_1 = tl_x_1 + patch_width
                    br_y_1 = tl_y_1 + patch_height
                    patch_area = (br_x_1 - tl_x_1) * (br_y_1 - tl_y_1)
                    if patch_area > (max_area_to_modify / 2) and not patch_coords:
                        continue
                    patch_coords.append([tl_x_1, tl_y_1])
                    patch_coords.append([br_x_1, br_y_1])

                    # Get coordinates of second patch, ensuring the second
                    # patch will not overlap with the first patch
                    tl_y_2 = np.random.randint(img_height - patch_height)
                    if tl_y_2 <= tl_y_1 - patch_height or tl_y_2 >= br_y_1:
                        tl_x_2 = np.random.randint(img_width - patch_width)
                    else:
                        possible_columns = list(range(tl_x_1 - patch_width + 1)) + list(
                            range(br_x_1, img_width - patch_width)
                        )
                        tl_x_2 = np.random.choice(np.array(possible_columns))
                    br_x_2 = tl_x_2 + patch_width
                    br_y_2 = tl_y_2 + patch_height
                    patch_coords.append([tl_x_2, tl_y_2])
                    patch_coords.append([br_x_2, br_y_2])

                    # Record the areas of image that were modified by the current
                    # pair of patches
                    area_check[tl_y_1:br_y_1, tl_x_1:br_x_1] = 1
                    area_check[tl_y_2:br_y_2, tl_x_2:br_x_2] = 1
                coord_matrix = np.array(patch_coords).T
                self.precomputed_masks.append(coord_matrix)
            logger.info("Finished precomputing masks!")

    def get_transform(self, img: np.ndarray) -> MedTransform:
        """
        Determines the pairs of patches that will be swapped in the image.

        Args:
            img: A N x H x W x C image, containing N images with height H,
                    width W, and consisting of C channels.
        Returns:
            An instance of the MedTransform "Swap2DPatches",
                initialized with the appropriate parameters based on the
                input image.
        """
        num_images = img.shape[0]
        img_height = img.shape[1]
        img_width = img.shape[2]
        img_center = np.array([[(img_width - 1) / 2], [(img_height - 1) / 2]]).astype("float64")
        patch_pairs = [[] for _ in range(num_images)]

        if self.sampling_pattern == "uniform":
            for i in range(num_images):
                # Randomly select one of the precomputed list of patch pairs
                rand_idx = np.random.randint(self.num_precompute)
                coord_matrix = self.precomputed_masks[rand_idx]
                coord_matrix = coord_matrix.astype("float64")

                # Create rotation matrix to randomly rotate coordinates by 0,
                # 90, 180, or 270 degrees counter-clockwise
                num_rotate = np.random.randint(4)
                rot_rad = (90 * num_rotate) * (np.pi / 180)
                rot_matrix = np.array(
                    [[np.cos(rot_rad), -np.sin(rot_rad)], [np.sin(rot_rad), np.cos(rot_rad)]]
                ).astype("float64")

                # Move all points such that the origin is the center of the image
                coord_matrix -= img_center

                # Apply rotation matrix
                rot_coord = np.dot(rot_matrix, coord_matrix)

                # Reset origin to top left corner of the image
                rot_coord += img_center

                # Set all points to be integers
                rot_coord = np.rint(rot_coord).astype(int)

                # Get top-left and bottom-right coordinates
                rot_coord = np.reshape(rot_coord, (2, rot_coord.shape[1] // 2, 2))
                tl = np.min(rot_coord, axis=2)
                br = np.max(rot_coord, axis=2)

                # Reshape tl and br to (2 x 2 x num_pairs)
                tl = np.reshape(tl, (2, tl.shape[1] // 2, 2))
                tl = np.transpose(tl, (0, 2, 1))
                br = np.reshape(br, (2, br.shape[1] // 2, 2))
                br = np.transpose(br, (0, 2, 1))

                patch_pairs[i].extend([tl, br])

        else:
            for i in range(num_images):
                # Randomly select one of the precomputed masks
                rand_idx = np.random.randint(self.num_precompute)
                pd_mask = self.precomputed_masks[rand_idx]

                # Randomly rotate by either 0, 90, 180, or 270 degrees
                # counter-clockwise
                num_rotate = np.random.randint(4)
                pd_mask_rot = np.rot90(pd_mask, k=num_rotate)

                # Get sample locations
                qy, qx = np.where(pd_mask_rot)
                all_locs = [(y, x) for y, x in zip(qy, qx)]
                np.random.shuffle(all_locs)

                # Get pairs of patches
                tl = np.zeros((2, 2, len(all_locs) // 2)).astype(int)
                br = np.zeros((2, 2, len(all_locs) // 2)).astype(int)
                half_patch = self.max_height // 2
                for pair_idx in range(len(all_locs) // 2):
                    p1_qy = all_locs[pair_idx * 2][0]
                    p1_qx = all_locs[pair_idx * 2][1]
                    tl[:, 0, pair_idx] = [p1_qx - half_patch, p1_qy - half_patch]
                    br[:, 0, pair_idx] = [p1_qx + half_patch, p1_qy + half_patch]

                    p2_qy = all_locs[pair_idx * 2 + 1][0]
                    p2_qx = all_locs[pair_idx * 2 + 1][1]
                    tl[:, 1, pair_idx] = [p2_qx - half_patch, p2_qy - half_patch]
                    br[:, 1, pair_idx] = [p2_qx + half_patch, p2_qy + half_patch]
                patch_pairs[i].extend([tl, br])

        return Swap2DPatches(patch_pairs=patch_pairs)


def apply_transform_gens(
    transform_gens: Sequence[Union[TransformGen, MedTransform]], img: np.ndarray
) -> Tuple[np.ndarray, TransformList]:
    """
    Apply a list of :class:`TransformGen` or :class:`MedTransform` on the
    input images, and returns the transformed images and a list of transforms.

    We cannot simply create and return all transforms without
    applying it to the images, because a subsequent transform may
    need the output of the previous one.

    Args:
        transform_gens: A list of :class:`TransformGen` or
                        :class:`MedTransform` instances to be applied.
        img: A stack of images of type uint8 or floating point with 1 or 3
                channels. The shape of img will be (N x H x W x C). Therefore,
                each of the N images will have height, H, and width, W,
                with C channels.

    Returns:
        transformed_imgs (np.ndarray): The transformed images.
        transform_list (TransformList): A TransformList that contains the
                                        transforms that are used.
    """
    for g in transform_gens:
        assert isinstance(g, (TransformGen, MedTransform)), g

    check_dtype(img)

    tfms = []
    for g in transform_gens:
        tfm = g.get_transform(img) if isinstance(g, TransformGen) else g
        assert isinstance(
            tfm, MedTransform
        ), "TransformGen {} must return an instance of MedTransform! " "Got {} instead".format(
            g, tfm
        )
        img = tfm.apply_image(img)
        tfms.append(tfm)
    return img, TransformList(tfms)


def build_preprocessing(cfg: Config):
    transforms = []
    assert len(cfg.PREPROCESSING) == len(cfg.PREPROCESSING_ARGS), (
        "cfg.PREPROCESSING and cfg.PREPROCESSING_ARGS must have the same " "length"
    )
    for i, pp in enumerate(cfg.PREPROCESSING):
        transform_gen_module = importlib.import_module(__name__)
        transform_module = importlib.import_module("medsegpy.data.transforms.transform")
        try:
            transform_class = getattr(transform_gen_module, pp)
        except AttributeError:
            try:
                transform_class = getattr(transform_module, pp)
            except AttributeError:
                raise ValueError("{} is not a valid transform!".format(pp))
        assert inspect.isclass(transform_class), "{} is not a class!".format(pp)
        assert issubclass(transform_class, TransformGen) or issubclass(
            transform_class, MedTransform
        ), "{} is not a valid transform!".format(pp)
        transforms.append(transform_class(**cfg.PREPROCESSING_ARGS[i]))

    return transforms



# Implementation of refine module used by RefineNet
# (http://openaccess.thecvf.com/content_cvpr_2017/papers/Lin_RefineNet_Multi-Path_Refinement_CVPR_2017_paper.pdf)
from typing import List

from keras.layers import Activation, Add
from keras.layers import BatchNormalization as BN
from keras.layers import Conv2D, Layer, MaxPooling2D

from medsegpy.modeling.deeplab_2d.deeplab_model import BilinearUpsampling


def refine_module(xs_in: List[Layer], num_filters_in: int, num_filters_out: int, name_prefix: str):
    # 1. adaptive convolutions
    adaptive_convs = []

    for i in range(len(xs_in)):
        x = xs_in[i]
        adaptive_convs.append(
            __residual_conv_unit__(x, num_filters_in, "%s_adp-conv%d" % (name_prefix, i))
        )

    # 2. multi-resolution fusion
    x = __multi_resolution_fusion__(adaptive_convs, name_prefix)

    # 3. chained residual pooling
    x = __chained_residual_pooling__(x, num_filters_in, name_prefix)
    x = __residual_conv_unit__(x, num_filters_in, name_prefix)

    x = Conv2D(
        filters=num_filters_out, kernel_size=(3, 3), padding="same", name="%s_conv" % name_prefix
    )(x)
    x = BN(axis=-1, momentum=0.95, epsilon=0.001, name="%s_bn" % name_prefix)(x)
    x = Activation("relu", name="%s_relu" % name_prefix)(x)

    return x


def __residual_conv_unit__(x_in: Layer, num_filters: int, name_prefix: str):
    name_prefix = "%s_rcu" % name_prefix
    x = x_in
    for i in range(2):
        x = BN(axis=-1, momentum=0.95, epsilon=0.001, name="%s_bn_%d" % (name_prefix, i))(x)
        x = Activation("relu", name="%s_relu_%d" % (name_prefix, i))(x)
        x = Conv2D(
            filters=num_filters,
            kernel_size=(3, 3),
            padding="same",
            name="%s_conv_%d" % (name_prefix, i),
        )(x)

    x = Add(name="%s_add" % name_prefix)([x_in, x])
    return x


def __multi_resolution_fusion__(xs_in: List[Layer], name_prefix: str):
    name_prefix = "%s_mrf" % name_prefix
    if len(xs_in) == 1:
        return xs_in[0]

    input_shapes = []
    num_channels = []
    for x in xs_in:
        input_shapes.append(x.shape.as_list()[1:-1])
        num_channels.append(x.shape.as_list()[-1])

    # upsample all inputs to the largest input size
    # assume largest input size is larger in all dimensions
    max_input_shape = max(input_shapes, key=lambda x: x[0])
    largest_input_ind = input_shapes.index(max_input_shape)

    # smallest number of filters
    num_filters = min(num_channels)

    xs = []
    for i in range(len(xs_in)):
        x = xs_in[i]
        x = Conv2D(
            filters=num_filters,
            kernel_size=(3, 3),
            padding="same",
            name="%s_conv_%d" % (name_prefix, i),
        )(x)
        x = BN(axis=-1, momentum=0.95, epsilon=0.001, name="%s_bn_%d" % (name_prefix, i))(x)

        if i != largest_input_ind:
            x = BilinearUpsampling(
                output_size=max_input_shape, name="%s_blu_%d" % (name_prefix, i)
            )(x)
        xs.append(x)

    x = Add(name="%s_mrf_add" % name_prefix)(xs)
    return x


def __chained_residual_pooling__(x_in: Layer, num_filters: int, name_prefix: str):
    name_prefix = "%s_crp" % name_prefix
    x = BN(axis=-1, momentum=0.95, epsilon=0.001, name="%s_bn" % name_prefix)(x_in)
    x = Activation("relu", name="%s_relu" % name_prefix)(x)
    residuals = [x]
    for i in range(3):
        x = MaxPooling2D(
            pool_size=(3, 3), strides=1, padding="same", name="%s_pool_%d" % (name_prefix, i)
        )(x)
        x = Conv2D(
            filters=num_filters,
            kernel_size=(3, 3),
            padding="same",
            name="%s_conv_%d" % (name_prefix, i),
        )(x)
        residuals.append(x)

    x = Add(name="%s_add" % name_prefix)(residuals)

    return x



import logging
import os
import warnings

import numpy as np
from keras.initializers import he_normal
from keras.layers import BatchNormalization as BN
from keras.layers import Concatenate, Conv2D, Conv2DTranspose, Dropout, Input, MaxPooling2D
from keras.utils.layer_utils import get_source_inputs

from ..model import Model

logger = logging.getLogger(__name__)

# List of tissues that can be segmented
FEMORAL_CARTILAGE_STR = "fc"
MENISCUS_STR = "men"
PATELLAR_CARTILAGE_STR = "pc"
TIBIAL_CARTILAGE_STR = "tc"

# Absolute directory where this file lives
__ABS_DIR__ = os.path.dirname(os.path.abspath(__file__))

WEIGHTS_DICT = {
    FEMORAL_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_fc_weights--0.8968.h5"),
    MENISCUS_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_men_weights--0.7692.h5"),
    PATELLAR_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_pc_weights--0.6206.h5"),
    TIBIAL_CARTILAGE_STR: os.path.join(__ABS_DIR__, "weights/unet_2d_tc_weights--0.8625.h5"),
}

# Input size that is expected
# All inputs must be at least this size
DEFAULT_INPUT_SIZE = (288, 288, 1)


def unet_2d_model(input_size=DEFAULT_INPUT_SIZE, input_tensor=None, output_mode=None):
    """Generate Unet 2D model compatible with Keras 2

    :param input_size: tuple of input size - format: (height, width, 1)

    :rtype: Keras model

    :raise ValueError if input_size is not tuple or dimensions of input_size do not match
        (height, width, 1)
    """
    warnings.warn(
        "unet_2d_model is deprecated. Use `UNet2D(cfg).build_model()`.", DeprecationWarning
    )
    from medsegpy import glob_constants

    logger.info("Initializing unet with seed: %s" % str(glob_constants.SEED))
    SEED = glob_constants.SEED
    if input_tensor is None and (type(input_size) is not tuple or len(input_size) != 3):
        raise ValueError("input_size must be a tuple of size (height, width, 1)")

    nfeatures = [2**feat * 32 for feat in np.arange(6)]
    depth = len(nfeatures)

    conv_ptr = []

    # input layer
    inputs = input_tensor if input_tensor is not None else Input(input_size)

    # step down convolutional layers
    pool = inputs
    for depth_cnt in range(depth):

        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(pool)
        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.0)(conv)

        conv_ptr.append(conv)

        # Only maxpool till penultimate depth
        if depth_cnt < depth - 1:

            # If size of input is odd, only do a 3x3 max pool
            xres = conv.shape.as_list()[1]
            if xres % 2 == 0:
                pooling_size = (2, 2)
            elif xres % 2 == 1:
                pooling_size = (3, 3)

            pool = MaxPooling2D(pool_size=pooling_size)(conv)

    # step up convolutional layers
    for depth_cnt in range(depth - 2, -1, -1):

        deconv_shape = conv_ptr[depth_cnt].shape.as_list()
        deconv_shape[0] = None

        # If size of input is odd, then do a 3x3 deconv
        if deconv_shape[1] % 2 == 0:
            unpooling_size = (2, 2)
        elif deconv_shape[1] % 2 == 1:
            unpooling_size = (3, 3)

        up = Concatenate(axis=3)(
            [
                Conv2DTranspose(
                    nfeatures[depth_cnt], (3, 3), padding="same", strides=unpooling_size
                )(conv),
                conv_ptr[depth_cnt],
            ]
        )

        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(up)
        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.00)(conv)

    # combine features
    # this if statement is required for legacy purposes
    # some weights were trained with a joint activation,
    # which makes it difficult to load weights effectively
    if output_mode is not None:
        recon = Conv2D(
            1,
            (1, 1),
            padding="same",
            activation=output_mode,
            kernel_initializer=he_normal(seed=SEED),
        )(conv)
    else:
        recon = conv

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)

    model = Model(inputs=inputs, outputs=[recon])

    return model


def unet_2d_model_v2(
    input_size=DEFAULT_INPUT_SIZE, input_tensor=None, output_mode=None, num_filters=None, depth=6
):
    """Generate Unet 2D model compatible with Keras 2

    :param input_size: tuple of input size - format: (height, width, 1)

    :rtype: Keras model

    :raise ValueError if input_size is not tuple or dimensions of input_size do not match
        (height, width, 1)
    """
    warnings.warn(
        "unet_2d_model_v2 is deprecated. Use `UNet2D(cfg).build_model()`.", DeprecationWarning
    )
    from medsegpy import glob_constants

    logger.info("Initializing unet with seed: %s" % str(glob_constants.SEED))
    SEED = glob_constants.SEED
    if input_tensor is None and (type(input_size) is not tuple or len(input_size) != 3):
        raise ValueError("input_size must be a tuple of size (height, width, 1)")

    if num_filters is None:
        nfeatures = [2**feat * 32 for feat in np.arange(depth)]
    else:
        nfeatures = num_filters
        assert len(nfeatures) == depth

    conv_ptr = []

    # input layer
    inputs = input_tensor if input_tensor is not None else Input(input_size)

    # step down convolutional layers
    pool = inputs
    for depth_cnt in range(depth):

        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(pool)
        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.0)(conv)

        conv_ptr.append(conv)

        # Only maxpool till penultimate depth
        if depth_cnt < depth - 1:

            # If size of input is odd, only do a 3x3 max pool
            xres = conv.shape.as_list()[1]
            if xres % 2 == 0:
                pooling_size = (2, 2)
            elif xres % 2 == 1:
                pooling_size = (3, 3)

            pool = MaxPooling2D(pool_size=pooling_size)(conv)

    # step up convolutional layers
    for depth_cnt in range(depth - 2, -1, -1):

        deconv_shape = conv_ptr[depth_cnt].shape.as_list()
        deconv_shape[0] = None

        # If size of input is odd, then do a 3x3 deconv
        if deconv_shape[1] % 2 == 0:
            unpooling_size = (2, 2)
        elif deconv_shape[1] % 2 == 1:
            unpooling_size = (3, 3)

        up = Concatenate(axis=3)(
            [
                Conv2DTranspose(
                    nfeatures[depth_cnt],
                    (3, 3),
                    padding="same",
                    strides=unpooling_size,
                    kernel_initializer=he_normal(seed=SEED),
                )(conv),
                conv_ptr[depth_cnt],
            ]
        )

        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(up)
        conv = Conv2D(
            nfeatures[depth_cnt],
            (3, 3),
            padding="same",
            activation="relu",
            kernel_initializer=he_normal(seed=SEED),
        )(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.00)(conv)

    # combine features
    # this if statement is required for legacy purposes
    # some weights were trained with a joint activation,
    # which makes it difficult to load weights effectively
    if output_mode is not None:
        recon = Conv2D(
            1,
            (1, 1),
            padding="same",
            activation=output_mode,
            kernel_initializer=he_normal(seed=SEED),
        )(conv)
    else:
        recon = conv

    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)

    model = Model(inputs=inputs, outputs=[recon])

    return model



from typing import Dict, Sequence, Union

import numpy as np
import tensorflow as tf
from keras import Input
from keras.layers import BatchNormalization as BN
from keras.layers import (
    Concatenate,
    Conv2D,
    Conv2DTranspose,
    Conv3D,
    Conv3DTranspose,
    Dropout,
    MaxPooling2D,
    MaxPooling3D,
)
from keras.utils.layer_utils import get_source_inputs

from medsegpy.config import UNet3DConfig, UNetConfig
from medsegpy.modeling.layers.attention import (
    CreateGatingSignal2D,
    CreateGatingSignal3D,
    DeepSupervision2D,
    DeepSupervision3D,
    MultiAttentionModule2D,
    MultiAttentionModule3D,
)
from medsegpy.modeling.model_utils import add_sem_seg_activation, zero_pad_like

from ..model import Model
from .build import META_ARCH_REGISTRY, ModelBuilder


def build_encoder_block(
    x: tf.Tensor,
    num_filters: Union[int, Sequence[int]],
    kernel_size: Union[int, Sequence[int]] = 3,
    num_conv: int = 2,
    activation: str = "relu",
    kernel_initializer: Union[str, Dict] = "he_normal",
    dropout: float = 0.0,
):
    """Builds simple FCN encoder block.

    Structure is below. Where blocks in `[]` are repeated:

    [Conv -> Activation] -> BN -> Dropout.

    Args:
        x (tf.Tensor): Input tensor.
        num_filters (int or Sequence[int]): Number of filters to use
            for each conv layer. If a sequence, will override number of conv
            layers specified by `num_conv`.
        kernel_size: Kernel size accepted by Keras convolution layers.
        num_conv (int, optional): Number of convolutional blocks
            (conv + activation) to use.
        activation (str, optional): Activation type.
        kernel_initializer: Kernel initializer accepted by
            `keras.layers.Conv(...)`.
        dropout (float, optional): Dropout rate.

    Returns:
        tf.Tensor: Encoder block output.
    """
    if len(x.shape) == 4:  # 2D
        conv_type = Conv2D
    elif len(x.shape) == 5:  # 3D
        conv_type = Conv3D
    else:
        raise ValueError("Only 2D or 3D inputs are supported")

    if isinstance(num_filters, int):
        num_filters = [num_filters] * num_conv

    for filters in num_filters:
        x = conv_type(
            filters,
            kernel_size,
            padding="same",
            activation=activation,
            kernel_initializer=kernel_initializer,
        )(x)
    x = BN(axis=-1, momentum=0.95, epsilon=0.001)(x)
    x = Dropout(rate=dropout)(x)
    return x


def build_decoder_block(
    x,
    x_skip,
    num_filters: Union[int, Sequence[int]],
    unpool_size: Sequence[int],
    kernel_size: Union[int, Sequence[int]] = 3,
    num_conv: int = 2,
    activation: str = "relu",
    kernel_initializer: Union[str, Dict] = "he_normal",
    dropout: float = 0.0,
):
    if len(x.shape) == 4:  # 2D
        conv_transpose_type = Conv2DTranspose
    elif len(x.shape) == 5:  # 3D
        conv_transpose_type = Conv3DTranspose
    else:
        raise ValueError("Only 2D or 3D inputs are supported")

    if isinstance(num_filters, int):
        # conv_transpose + num_conv
        num_filters = [num_filters] * (num_conv + 1)

    x_shape = x.shape.as_list()[1:-1]
    x = conv_transpose_type(
        num_filters[0],
        kernel_size,
        padding="same",
        strides=unpool_size,
        kernel_initializer=kernel_initializer,
    )(x)
    x_shape = [x * y for x, y in zip(x_shape, unpool_size)]
    x = Concatenate(axis=-1)([zero_pad_like(x, x_skip, x_shape), x_skip])

    if len(num_filters) > 1:
        x = build_encoder_block(
            x, num_filters[1:], kernel_size, num_conv, activation, kernel_initializer, dropout
        )
    return x


@META_ARCH_REGISTRY.register()
class UNet2D(ModelBuilder):
    def __init__(self, cfg: UNetConfig):
        super().__init__(cfg)
        self._pooler_type = MaxPooling2D
        self._conv_type = Conv2D
        self._multi_attention_module = MultiAttentionModule2D
        self._create_gating_signal = CreateGatingSignal2D
        self._deep_supervision = DeepSupervision2D

        self._dim = 2
        self.kernel_size = (3, 3)
        self.use_attention = False
        self.use_deep_supervision = False

    def build_model(self, input_tensor=None) -> Model:
        cfg = self._cfg
        seed = cfg.SEED
        depth = cfg.DEPTH
        kernel_size = self.kernel_size
        self.use_attention = cfg.USE_ATTENTION
        self.use_deep_supervision = cfg.USE_DEEP_SUPERVISION

        kernel_initializer = {"class_name": cfg.KERNEL_INITIALIZER, "config": {"seed": seed}}
        output_mode = cfg.LOSS[1]
        assert output_mode in ["sigmoid", "softmax"]
        num_classes = cfg.get_num_classes()
        num_filters = cfg.NUM_FILTERS
        if not num_filters:
            num_filters = [2**feat * 32 for feat in range(depth)]
        else:
            depth = len(num_filters)

        # Build inputs.
        if not input_tensor:
            inputs = self._build_input(cfg.IMG_SIZE)
        else:
            inputs = input_tensor

        # Encoder.
        x_skips = []
        pool_sizes = []
        x = inputs
        deep_supervision_outputs = []
        scale_factors = np.ones(((depth - 1), self._dim), dtype=int)
        for depth_cnt in range(depth):
            x = build_encoder_block(
                x,
                num_filters[depth_cnt],
                kernel_size=kernel_size,
                num_conv=2,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=0.0,
            )

            # Maxpool until penultimate depth.
            if depth_cnt < depth - 1:
                x_skips.append(x)
                pool_size = self._get_pool_size(x)
                pool_sizes.append(pool_size)
                if depth_cnt < depth - 2:
                    scale_factors[depth_cnt + 1] = scale_factors[depth_cnt] * pool_size
                x = self._pooler_type(pool_size=pool_size)(x)

        # Decoder.
        for i, (x_skip, unpool_size) in enumerate(zip(x_skips[::-1], pool_sizes[::-1])):
            depth_cnt = depth - i - 2
            skip_connect = x_skip
            # The first skip connection is not passed through
            # an attention gate, as mentioned in the paper under
            # the section "Attention Gates in U-Net Model"
            if depth_cnt > 0 and self.use_attention:
                if i == 0:
                    gating_signal = self._create_gating_signal(
                        out_channels=num_filters[depth_cnt + 1]
                    )(x)
                else:
                    gating_signal = x
                attn_out, attn_coeffs = self._multi_attention_module(
                    in_channels=num_filters[depth_cnt],
                    intermediate_channels=num_filters[depth_cnt],
                    sub_sample_factor=self._get_pool_size(x_skip),
                )([x_skip, gating_signal])
                skip_connect = attn_out

            x = build_decoder_block(
                x,
                skip_connect,
                num_filters[depth_cnt],
                unpool_size,
                kernel_size=kernel_size,
                num_conv=2,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=0.0,
            )

            if self.use_deep_supervision:
                # Determine scale factor for deep supervision
                deep_supervision_outputs.append(
                    self._deep_supervision(
                        out_channels=num_classes, scale_factor=tuple(scale_factors[depth_cnt])
                    )(x)
                )

        if self.use_deep_supervision:
            x = Concatenate(axis=-1)(deep_supervision_outputs)

        # 1x1 convolution to get pixel-wise semantic segmentation.
        x = add_sem_seg_activation(
            x,
            num_classes,
            conv_type=self._conv_type,
            activation=output_mode,
            kernel_initializer="glorot_uniform",
            seed=seed,
        )

        if input_tensor is not None:
            inputs = get_source_inputs(input_tensor)

        model = Model(inputs=inputs, outputs=[x])

        return model

    def _build_input(self, input_size):
        if len(input_size) == 2:
            input_size = input_size + (1,)
        elif len(input_size) != 3:
            raise ValueError("2D U-Net must have an input of shape HxWxC")

        return Input(input_size)

    def _get_pool_size(self, x: tf.Tensor):
        return [2 if d % 2 == 0 or d % 3 != 0 else 3 for d in x.shape.as_list()[1:-1]]


@META_ARCH_REGISTRY.register()
class UNet3D(UNet2D):
    def __init__(self, cfg: UNet3DConfig):
        super().__init__(cfg)
        self._pooler_type = MaxPooling3D
        self._conv_type = Conv3D
        self._multi_attention_module = MultiAttentionModule3D
        self._create_gating_signal = CreateGatingSignal3D
        self._deep_supervision = DeepSupervision3D

        self._dim = 3
        self.kernel_size = (3, 3, 3)

    def _build_input(self, input_size):
        if len(input_size) == 3:
            input_size = input_size + (1,)
        elif len(input_size) != 4:
            raise ValueError("2D U-Net must have an input of shape HxWxDxC")

        return Input(input_size)

    def _get_pool_size(self, x: tf.Tensor):
        return (2, 2, 1) if x.shape.as_list()[-2] // 2 == 0 else (2, 2, 2)



import re
from abc import ABC, abstractmethod


class FnameParser(ABC):
    """Abstract class for parsing filenames into comprehensible information"""

    FNAME_REGEX = ""

    @abstractmethod
    def get_file_info(self, fname_or_filepath: str) -> dict:
        """Returns dictionary containing parsed file information

        Dictionary must contain the following keys:
            * "pid" (int): patient/subject id
            * "aug" (int): augmentation id
            * "slice" (int): Slice number
            * "scan_id" (str): Scan id
            * "volume_id" (str): Volume id - scan_id + augmentation id

        Args:
            fname_or_filepath (str): Filename or filepath.

        Returns:
            dict: File information extracted from filename.
        """
        pass

    @abstractmethod
    def get_file_id(self, fname: str):
        pass

    @abstractmethod
    def get_fname(self, file_info: dict):
        pass

    @abstractmethod
    def get_pid_from_volume_id(self, volume_id: str):
        pass


class OAISliceWise(FnameParser):
    """Parse information from files stored in the medsegpy 2D filename format.

    See datasets/README.md for more information.
    """

    # sample fname: 9311328_V01-Aug04_072.im - format: %7d_V%02d-Aug%02d_%03d
    FNAME_FORMAT = "%07d_V%02d-Aug%02d_%03d"
    FNAME_REGEX = "([\d]+)_V([\d]+)-Aug([\d]+)_([\d]+)"
    _VOLUME_ID_FORMAT = "%d_V%02d-Aug%02d"

    def get_file_info(self, fname_or_filepath: str) -> dict:
        _, pid, timepoint, augmentation, slice_num, _ = tuple(
            re.split(self.FNAME_REGEX, fname_or_filepath)
        )
        pid = int(pid)
        timepoint = int(timepoint)
        augmentation = int(augmentation)
        slice_num = int(slice_num)

        return {
            "pid": pid,
            "timepoint": timepoint,
            "aug": augmentation,
            "slice": slice_num,
            "scanid": "%d_V%02d" % (pid, timepoint),
            "volume_id": self._VOLUME_ID_FORMAT % (pid, timepoint, augmentation),
        }

    def get_file_id(self, fname):
        fname_info = self.get_file_info(fname)
        return (
            str(fname_info["pid"])
            + str(fname_info["timepoint"])
            + str(fname_info["aug"])
            + str(fname_info["slice"])
        )

    def get_fname(self, file_info: dict):
        return self.FNAME_FORMAT % (
            file_info["pid"],
            file_info["timepoint"],
            file_info["aug"],
            file_info["slice"],
        )

    def get_pid_from_volume_id(self, volume_id: str):
        return int(volume_id.split("_")[0])



"""Stanford qDESS dataset"""
import json
import os

import dosma as dm
import h5py
import numpy as np
import pandas as pd
from dosma.scan_sequences.mri import QDess
from dosma.tissues import FemoralCartilage, Meniscus, PatellarCartilage, TibialCartilage

from medsegpy.data.datasets.qdess_mri import QDESS_SEGMENTATION_CATEGORIES
from medsegpy.evaluation import EVALUATOR_REGISTRY, SemSegEvaluator
from medsegpy.evaluation.metrics import Metric, Reductions, flatten_non_category_dims


class T2Relaxation(Metric):
    """Compute T2 relaxation time for each tissue."""

    _METADATA_PATH = "/dataNAS/people/arjun/data/skm-tea/v1-release/all_metadata.csv"
    _BASE_SCAN_DIR = "/dataNAS/people/arjun/data/skm-tea/v1-release/image_files"
    _TEST_ANNOTATION_PATH = "/bmrNAS/people/arjun/data/qdess_knee_2020/annotations/v0.0.1/test.json"

    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["pred", "base"]
        self._method = method

        # Get specific image metadata for all scans in the test set
        with open(self._TEST_ANNOTATION_PATH, "r") as fp:
            test_metadata = json.load(fp)
        self.scan_to_metadata = {
            img_dict["scan_id"]: img_dict for img_dict in test_metadata["images"]
        }

        # Get additional metadata for all scans
        self.additional_metadata = pd.read_csv(self._METADATA_PATH)

    def _compute_t2_relaxation(self, y, scan_id, category_dim: int = None):
        # Convert mask to binary
        y_bin = y.astype(np.bool)

        # Get scan metadata
        scan_metadata = self.scan_to_metadata[scan_id]

        # Build affine matrix
        affine = dm.to_affine(scan_metadata["orientation"], scan_metadata["voxel_spacing"])

        # Make qDESS object
        scan_path = os.path.join(self._BASE_SCAN_DIR, f"{scan_id}.h5")
        with h5py.File(scan_path, "r") as fp:
            e1 = dm.MedicalVolume(fp["echo1"][()], affine)
            e2 = dm.MedicalVolume(fp["echo2"][()], affine)
        qdess = QDess([e1, e2])

        # Determine if left or right side corresponds to the medial direction
        scan_additional_metadata = self.additional_metadata[
            self.additional_metadata["MTR_ID"] == scan_id
        ]
        is_left_medial = (scan_additional_metadata["MedialDirection"] == "L").item()

        # Make tissue objects
        fc = FemoralCartilage(medial_to_lateral=is_left_medial)
        tc = TibialCartilage(medial_to_lateral=is_left_medial)
        pc = PatellarCartilage(medial_to_lateral=is_left_medial)
        men = Meniscus(medial_to_lateral=is_left_medial)

        # Set mask for each tissue
        category_indices = {x["abbrev"]: idx for idx, x in enumerate(QDESS_SEGMENTATION_CATEGORIES)}
        fc.set_mask(
            dm.MedicalVolume(y_bin[..., category_indices["fc"]], affine), use_largest_cc=True
        )
        tc.set_mask(
            dm.MedicalVolume(y_bin[..., category_indices["tc"]], affine), use_largest_ccs=True
        )
        pc.set_mask(
            dm.MedicalVolume(y_bin[..., category_indices["pc"]], affine), use_largest_cc=True
        )
        men.set_mask(
            dm.MedicalVolume(y_bin[..., category_indices["men"]], affine), use_largest_ccs=True
        )

        # Calculate T2 relaxation time for femoral cartilage
        fc_t2_map = qdess.generate_t2_map(
            tissue=fc,
            suppress_fat=True,
            suppress_fluid=True,
            gl_area=float(scan_additional_metadata["SpoilerGradientArea"]),
            tg=float(scan_additional_metadata["SpoilerGradientTime"]),
            tr=float(scan_additional_metadata["RepetitionTime"]),
            te=float(scan_additional_metadata["EchoTime1"]),
            alpha=float(scan_additional_metadata["FlipAngle"]),
            t1=1200,
            nan_bounds=(0, 100),
        )

        # Calculate T2 relaxation time for tibial cartilage
        tc_t2_map = qdess.generate_t2_map(
            tissue=tc,
            suppress_fat=True,
            suppress_fluid=True,
            gl_area=float(scan_additional_metadata["SpoilerGradientArea"]),
            tg=float(scan_additional_metadata["SpoilerGradientTime"]),
            tr=float(scan_additional_metadata["RepetitionTime"]),
            te=float(scan_additional_metadata["EchoTime1"]),
            alpha=float(scan_additional_metadata["FlipAngle"]),
            t1=1200,
            nan_bounds=(0, 100),
        )

        # Calculate T2 relaxation time for patellar cartilage
        pc_t2_map = qdess.generate_t2_map(
            tissue=pc,
            suppress_fat=True,
            suppress_fluid=True,
            gl_area=float(scan_additional_metadata["SpoilerGradientArea"]),
            tg=float(scan_additional_metadata["SpoilerGradientTime"]),
            tr=float(scan_additional_metadata["RepetitionTime"]),
            te=float(scan_additional_metadata["EchoTime1"]),
            alpha=float(scan_additional_metadata["FlipAngle"]),
            t1=1200,
            nan_bounds=(0, 100),
        )

        # Calculate T2 relaxation time for meniscus
        men_t2_map = qdess.generate_t2_map(
            tissue=men,
            suppress_fat=True,
            suppress_fluid=True,
            gl_area=float(scan_additional_metadata["SpoilerGradientArea"]),
            tg=float(scan_additional_metadata["SpoilerGradientTime"]),
            tr=float(scan_additional_metadata["RepetitionTime"]),
            te=float(scan_additional_metadata["EchoTime1"]),
            alpha=float(scan_additional_metadata["FlipAngle"]),
            t1=1200,
            nan_bounds=(0, 100),
        )

        # Return mean T2 relaxation time for each tissue
        mean_t2_vals = [0.0, 0.0, 0.0, 0.0]
        mean_t2_vals[category_indices["fc"]] = np.mean(
            fc_t2_map.volumetric_map.volume[y_bin[..., category_indices["fc"]]]
        )
        mean_t2_vals[category_indices["tc"]] = np.mean(
            tc_t2_map.volumetric_map.volume[y_bin[..., category_indices["tc"]]]
        )
        mean_t2_vals[category_indices["pc"]] = np.mean(
            pc_t2_map.volumetric_map.volume[y_bin[..., category_indices["pc"]]]
        )
        mean_t2_vals[category_indices["men"]] = np.mean(
            men_t2_map.volumetric_map.volume[y_bin[..., category_indices["men"]]]
        )

        return np.array(mean_t2_vals)

    def __call__(self, y_pred, y_true, scan_id, category_dim: int = None):
        # Ensure category_dim is -1
        assert category_dim == -1

        # Compute T2 relaxation time
        if self._method == "pred":
            t2_time = self._compute_t2_relaxation(y_pred, scan_id, category_dim)
        else:
            t2_time = self._compute_t2_relaxation(y_true, scan_id, category_dim)

        return (t2_time,)

    def name(self):
        return f"T2 {self._method}"


class TissueVolume(Metric):
    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["pred", "base"]
        self._method = method

    def _compute_volume(self, y, pixel_volume, category_dim):
        y = flatten_non_category_dims(y.astype(np.bool), category_dim)
        return pixel_volume * np.count_nonzero(y, -1)

    def __call__(self, y_pred, y_true, spacing=None, category_dim: int = None):
        # Ensure category_dim is -1
        assert category_dim == -1

        # Get volume of each pixel
        pixel_volume = np.prod(spacing) if spacing else 1

        # Compute volume of tissue
        if self._method == "pred":
            tissue_volume = self._compute_volume(y_pred, pixel_volume, category_dim)
        else:
            tissue_volume = self._compute_volume(y_true, pixel_volume, category_dim)

        return (tissue_volume,)

    def name(self):
        return f"Volume {self._method}"


class T2RelaxationDiff(Metric):
    """Compare T2 Relaxation times using predicted segmentation
    and ground truth segmentation
    """

    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["absolute", "percent"]
        self._method = method

        self._t2_pred = T2Relaxation(method="pred")
        self._t2_true = T2Relaxation(method="base")

    def __call__(self, y_pred, y_true, scan_id, category_dim: int = None):
        # Ensure category_dim is -1
        assert category_dim == -1

        # Calculate T2 relaxation times
        t2_time_pred = self._t2_pred(y_pred, y_true, scan_id, category_dim)
        t2_time_true = self._t2_true(y_pred, y_true, scan_id, category_dim)

        # Determine difference
        vals = t2_time_pred[0] - t2_time_true[0]
        if self._method == "percent":
            vals = (vals / t2_time_true[0]) * 100

        return (vals,)

    def name(self):
        symbol = "Abs" if self._method == "absolute" else "%"
        return f"T2 {symbol}Diff"


class TissueVolumeDiff(Metric):
    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["absolute", "percent"]
        self._method = method

        self._volume_pred = TissueVolume(method="pred", units="(mm^3)")
        self._volume_true = TissueVolume(method="base", units="(mm^3)")

    def __call__(self, y_pred, y_true, spacing=None, category_dim: int = None):
        # Ensure category_dim is -1
        assert category_dim == -1

        # Calculate tissue volumes
        volume_pred = self._volume_pred(y_pred, y_true, spacing, category_dim)
        volume_true = self._volume_true(y_pred, y_true, spacing, category_dim)

        # Determine difference
        vals = volume_pred[0] - volume_true[0]
        if self._method == "percent":
            vals = (vals / volume_true[0]) * 100

        return (vals,)

    def name(self):
        symbol = "Abs" if self._method == "absolute" else "%"
        return f"Volume {symbol}Diff"


@EVALUATOR_REGISTRY.register()
class QDESSEvaluator(SemSegEvaluator):
    """Evaluate semantic segmentation on the qDESS dataset."""

    def _get_metrics(self):
        metrics = list(super()._get_metrics())
        paired_metrics = [
            T2Relaxation(method="pred"),
            T2Relaxation(method="base"),
            TissueVolume(method="pred", units="(mm^3)"),
            TissueVolume(method="base", units="(mm^3)"),
        ]
        metrics.extend(
            [
                T2RelaxationDiff(method="absolute"),
                T2RelaxationDiff(method="percent"),
                TissueVolumeDiff(method="absolute", units="(mm^3)"),
                TissueVolumeDiff(method="percent"),
            ]
        )
        metrics.extend(paired_metrics)

        self._metric_pairs = [
            (p.name(), b.name()) for p, b in zip(paired_metrics[::2], paired_metrics[1::2])
        ]
        return metrics

    def reset(self):
        super().reset()
        for pred_key, base_key in self._metric_pairs:
            self._metrics_manager.register_pairs(
                pred_key, base_key, (Reductions.RMS_CV, Reductions.RMSE_CV)
            )

    def process(self, inputs, outputs):
        if not self.spacing:
            assert all("scan_spacing" in x for x in inputs)
        super().process(inputs, outputs)



"""Dataset evaluator.

Adopted from Facebook's detectron2.
https://github.com/facebookresearch/detectron2
"""
import datetime
import logging
import time
from typing import Sequence, Union

from medsegpy.data import DataLoader
from medsegpy.data.im_gens import Generator, GeneratorState
from medsegpy.utils.logger import log_every_n_seconds


class DatasetEvaluator:
    """
    Base class for a dataset evaluator.

    The function :func:`inference_on_dataset` runs the model over
    all samples in the dataset, and have a DatasetEvaluator to process the
        inputs/outputs.

    This class will accumulate information of the inputs/outputs
        (by :meth:`process`),
    and produce evaluation results in the end (by :meth:`evaluate`).
    """

    def reset(self):
        """
        Preparation for a new round of evaluation.
        Should be called before starting a round of evaluation.
        """
        pass

    def process(self, inputs, outputs):
        """
        Process an input/output pair.

        Args:
            scan_id: the scan id corresponding to the input/output
            inputs (List[Dict]]: the inputs that are used to call the model.
                Can also contain scan specific fields. These fields
                should start with "scan_".
            outputs (List[Dict]): List of outputs from the model.
                Each dict should contain at least the following keys:
                * "y_true": Ground truth results
                * "y_pred": Predicted probabilities.
                * "time_elapsed": Amount of time to load data and run model.
        """
        pass

    def evaluate(self):
        """Evaluate/summarize the performance, after processing all input/output
        pairs.

        Returns:
            dict:
                A new evaluator class can return a dict of arbitrary format
                as long as the user can process the results.
                In our train_net.py, we expect the following format:

                * key: the name of the task (e.g., bbox)
                * value: a dict of {metric name: score}, e.g.: {"AP50": 80}
        """
        pass


def inference_on_dataset(
    model,
    data_loader: Union[DataLoader, Generator],
    evaluator: Union[DatasetEvaluator, Sequence[DatasetEvaluator]],
):
    """
    Run model on the data_loader and evaluate the metrics with evaluator.
    The model will be used in eval mode.

    Args:
        model (keras.Model):
        generator: an iterable object with a length.
            The elements it generates will be the inputs to the model.
        evaluator (DatasetEvaluator): the evaluator to run. Use
            :class:`DatasetEvaluators([])` if you only want to benchmark, but
            don't want to do any evaluation.

    Returns:
        The return value of `evaluator.evaluate()`
    """
    if isinstance(evaluator, DatasetEvaluator):
        evaluator = [evaluator]

    for e in evaluator:
        e.reset()
    num_warmup = 1
    start_time = time.perf_counter()
    total_compute_time = 0
    total_processing_time = 0
    total_inference_time = 0
    if isinstance(data_loader, Generator):
        iter_loader = data_loader.img_generator_test
        total = data_loader.num_scans(GeneratorState.TESTING)
    else:
        iter_loader = data_loader.inference
        total = data_loader.num_scans()

    start_compute_time = time.perf_counter()
    logger = logging.getLogger(__name__)
    for idx, (input, output) in enumerate(iter_loader(model)):
        total_compute_time += time.perf_counter() - start_compute_time

        start_processing_time = time.perf_counter()
        for e in evaluator:
            e.process([input], [output])
        total_processing_time += time.perf_counter() - start_processing_time

        total_inference_time += output["time_elapsed"]
        iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
        seconds_per_scan = total_compute_time / iters_after_start
        seconds_per_inference = total_inference_time / iters_after_start
        seconds_per_processing = total_processing_time / iters_after_start

        if idx >= num_warmup * 2 or seconds_per_scan > 5:
            total_seconds_per_img = (time.perf_counter() - start_time) / iters_after_start
            eta = datetime.timedelta(seconds=int(total_seconds_per_img * (total - idx - 1)))
            log_every_n_seconds(
                logging.INFO,
                "Inference done {}/{}. {:.4f} s / scan ({:.4f} inference, "
                "{:.4f} processing). ETA={}".format(
                    idx + 1,
                    total,
                    seconds_per_scan,
                    seconds_per_inference,
                    seconds_per_processing,
                    str(eta),
                ),
                n=5,
            )
        start_compute_time = time.perf_counter()

    eval_start = time.perf_counter()
    logger.info("Begin evaluation...")
    results = {e.__class__.__name__: e.evaluate() for e in evaluator}
    total_eval_time = time.perf_counter() - eval_start
    logger.info("Time Elapsed: {:.4f} seconds".format(total_compute_time + total_eval_time))
    # An evaluator may return None when not in main process.
    # Replace it by an empty dict instead to make it easier for downstream
    # code to handle
    if results is None:
        results = {}
    return results



import logging
from abc import ABC, abstractmethod

from keras.initializers import glorot_uniform, he_normal
from keras.layers import Input
from keras.layers.convolutional import Convolution2D
from keras.layers.core import Activation
from keras.layers.normalization import BatchNormalization

from ..layers import MaxPoolingWithArgmax2D, MaxUnpooling2D
from ..model import Model

logger = logging.getLogger(__name__)


class KModel(ABC):
    def __init__(self, input_shape=None, input_tensor=None, seed=None):
        self._input_shape = input_shape
        self._input_tensor = input_tensor
        self._seed = seed

    @abstractmethod
    def build_model(self):
        pass


class SegNet(KModel):
    def __init__(self, input_shape=None, input_tensor=None, seed=None, **kwargs):
        super().__init__(input_shape, input_tensor, seed)

        self._n_labels = kwargs.get("n_labels")
        self._depth = kwargs.get("depth")
        self._num_conv_layers = kwargs.get("num_conv_layers")
        self._num_filters = kwargs.get("num_filters")
        self._kernel = kwargs.get("kernel", 3)
        self._pool_size = kwargs.get("pool_size", (2, 2))
        self._output_mode = kwargs.get("output_mode")
        self._single_bn = kwargs.get("single_bn", False)
        self._conv_act_bn = kwargs.get("conv_act_bn", False)

    def build_model(self):
        input_tensor = self._input_tensor
        input_shape = self._input_shape
        seed = self._seed
        n_labels = self._n_labels
        depth = self._depth
        num_conv_layers = self._num_conv_layers
        num_filters = self._num_filters
        kernel = self._kernel
        pool_size = self._pool_size
        output_mode = self._output_mode
        single_bn = self._single_bn
        conv_act_bn = self._conv_act_bn

        logger.info("Initializing SegNet with seed: {}".format(self._seed))

        inputs = input_tensor if input_tensor else Input(shape=input_shape)

        mask_layers = []

        curr_layer = inputs
        eff_pool_sizes = []

        # Determine pool sizes
        for i in range(depth):
            level = i + 1
            eff_pool_size = pool_size
            divisor = pool_size[0] ** level
            if input_shape[0] % divisor != 0:
                eff_pool_size = (3, 3)
            eff_pool_sizes.append(eff_pool_size)

        # encoder
        logger.info("Building Encoder...")
        for i in range(depth):
            eff_pool_size = eff_pool_sizes[i]
            if conv_act_bn:
                curr_layer, l_mask = self._encoder_block_conv_act_bn(
                    curr_layer,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                )
            else:
                curr_layer, l_mask = self._encoder_block(
                    curr_layer,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                    single_bn=single_bn,
                )

            mask_layers.append(l_mask)

        logger.info("Building decoder...")
        # decoder
        for i in reversed(range(depth)):
            l_mask = mask_layers[i]
            eff_pool_size = eff_pool_sizes[i]
            if conv_act_bn:
                curr_layer = self._decoder_block_conv_act_bn(
                    curr_layer,
                    l_mask,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    num_filters_next=1 if i == 0 else num_filters[i - 1],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                )
            else:
                curr_layer = self._decoder_block(
                    curr_layer,
                    l_mask,
                    level=i + 1,
                    num_conv_layers=num_conv_layers[i],
                    num_filters=num_filters[i],
                    num_filters_next=num_filters[i] if i == 0 else num_filters[i - 1],
                    kernel=kernel,
                    pool_size=eff_pool_size,
                    single_bn=single_bn,
                )

        outputs = Convolution2D(
            n_labels, (1, 1), kernel_initializer=glorot_uniform(seed=seed), activation=output_mode
        )(curr_layer)

        segnet = Model(inputs=inputs, outputs=outputs, name="segnet")

        return segnet

    def _encoder_block(
        self,
        x,
        level,
        num_conv_layers=2,
        num_filters=64,
        kernel=3,
        pool_size=(2, 2),
        single_bn=False,
        add_pool: bool = True,
    ):
        seed = self._seed
        if num_conv_layers <= 0:
            raise ValueError("Must have at least 1 conv layer")

        curr_layer = x

        for i in range(num_conv_layers):
            conv = Convolution2D(
                num_filters,
                (kernel, kernel),
                padding="same",
                kernel_initializer=he_normal(seed=seed),
                name="enc_%d_conv_%d" % (level, i + 1),
            )(curr_layer)
            if not single_bn:
                conv = BatchNormalization(name="enc_%d_bn_%d" % (level, i + 1))(conv)
            conv = Activation("relu", name="enc_%d_relu_%d" % (level, i + 1))(conv)
            curr_layer = conv

        if single_bn:
            curr_layer = BatchNormalization(name="enc_%d_bn" % level)(curr_layer)

        if add_pool:
            l_pool, l_mask = MaxPoolingWithArgmax2D(
                pool_size=pool_size, strides=pool_size, name="enc_%d_pool" % level
            )(curr_layer)
        else:
            l_pool, l_mask = curr_layer, None

        return l_pool, l_mask

    def _decoder_block(
        self,
        x_pool,
        x_mask,
        level,
        num_conv_layers=2,
        num_filters=64,
        num_filters_next=32,
        kernel=3,
        pool_size=(2, 2),
        single_bn=False,
    ):
        seed = self._seed
        unpool_1 = MaxUnpooling2D(pool_size)([x_pool, x_mask])

        curr_layer = unpool_1

        for i in range(num_conv_layers):
            used_num_filters = num_filters_next if i == num_conv_layers - 1 else num_filters
            conv = Convolution2D(
                used_num_filters,
                (kernel, kernel),
                kernel_initializer=he_normal(seed=seed),
                padding="same",
                name="dec_%d_conv_%d" % (level, i + 1),
            )(curr_layer)

            if not single_bn:
                conv = BatchNormalization(name="dec_%d_bn_%d" % (level, i + 1))(conv)

            conv = Activation("relu", name="dec_%d_relu_%d" % (level, i + 1))(conv)
            curr_layer = conv

        if single_bn:
            curr_layer = BatchNormalization(name="dec_%d_bn" % level)(curr_layer)

        return curr_layer

    def _encoder_block_conv_act_bn(
        self,
        x,
        level,
        num_conv_layers=2,
        num_filters=64,
        kernel=3,
        pool_size=(2, 2),
        add_pool: bool = True,
    ):
        seed = self._seed

        if num_conv_layers <= 0:
            raise ValueError("Must have at least 1 conv layer")

        curr_layer = x

        for i in range(num_conv_layers):
            conv = Convolution2D(
                num_filters,
                (kernel, kernel),
                padding="same",
                kernel_initializer=he_normal(seed=seed),
                name="enc_%d_conv_%d" % (level, i + 1),
            )(curr_layer)
            conv = Activation("relu", name="enc_%d_relu_%d" % (level, i + 1))(conv)
            conv = BatchNormalization(name="enc_%d_bn_%d" % (level, i + 1))(conv)
            curr_layer = conv

        if add_pool:
            l_pool, l_mask = MaxPoolingWithArgmax2D(
                pool_size=pool_size, strides=pool_size, name="enc_%d_pool" % level
            )(curr_layer)
        else:
            l_pool, l_mask = curr_layer, None

        return l_pool, l_mask

    def _decoder_block_conv_act_bn(
        self,
        x_pool,
        x_mask,
        level,
        num_conv_layers=2,
        num_filters=64,
        num_filters_next=32,
        kernel=3,
        pool_size=(2, 2),
    ):
        seed = self._seed

        unpool_1 = MaxUnpooling2D(pool_size)([x_pool, x_mask])

        curr_layer = unpool_1

        for i in range(num_conv_layers):
            used_num_filters = num_filters_next if i == num_conv_layers - 1 else num_filters
            conv = Convolution2D(
                used_num_filters,
                (kernel, kernel),
                padding="same",
                kernel_initializer=he_normal(seed=seed),
                name="dec_%d_conv_%d" % (level, i + 1),
            )(curr_layer)
            conv = Activation("relu", name="dec_%d_relu_%d" % (level, i + 1))(conv)
            conv = BatchNormalization(name="dec_%d_bn_%d" % (level, i + 1))(conv)
            curr_layer = conv

        return curr_layer


def Segnet_v2(**kwargs):
    m = SegNet(**kwargs)
    return m.build_model()



"""Stanford abdominal CT Dataset."""
import os

import numpy as np

from medsegpy.evaluation import EVALUATOR_REGISTRY, SemSegEvaluator
from medsegpy.evaluation.metrics import Metric, Reductions, flatten_non_category_dims


class HU(Metric):
    """
    Calculates the mean Hounsfield Unit value.
    """

    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["pred", "base"]
        self.method = method

    def _compute_hu(self, x, y, category_dim: int = None):
        y = y.astype(np.bool)
        if category_dim is None:
            return np.mean(x[y])

        num_classes = y.shape[-1]
        assert category_dim == -1
        hu = np.array([np.mean(x[y[..., c]]) for c in range(num_classes)])

        return hu

    def __call__(self, x, y_pred=None, y_true=None, category_dim: int = None):
        if self.method == "pred":
            assert y_pred is not None
            hu = self._compute_hu(x, y_pred, category_dim)
        if self.method == "base":
            assert y_true is not None
            hu = self._compute_hu(x, y_true, category_dim)

        return (hu,)

    def name(self):
        return "HU {}".format(self.method)


class CSA(Metric):
    """
    Calculates the cross-sectional area.
    """

    def __init__(self, method, units=""):
        super().__init__(units)
        assert method in ["pred", "base"]
        self.method = method

    def _compute_area(self, y, pixel_area, category_dim=None):
        y = flatten_non_category_dims(y.astype(np.bool), category_dim)
        return pixel_area * np.count_nonzero(y, -1)

    def __call__(self, y_pred=None, y_true=None, spacing=None, category_dim: int = None):
        pixel_area = np.prod(spacing) if spacing else 1
        if self.method == "pred":
            assert y_pred is not None
            area = self._compute_area(y_pred, pixel_area, category_dim)
        if self.method == "base":
            assert y_true is not None
            area = self._compute_area(y_true, pixel_area, category_dim)

        return (area,)

    def name(self):
        return "Area {}".format(self.method)


class HUDiff(Metric):
    """
    Calculates the absolute difference or the percent error in Hounsfield units
    when compared to the ground truth.
    """

    FULL_NAME = "Hounsfield Unit - Difference"

    def __init__(self, units="", method="absolute"):
        super().__init__(units)
        assert method in ["absolute", "percent"]
        self._method = method

    def __call__(self, y_pred, y_true, x, category_dim: int = None):
        y_true = y_true.astype(np.bool)
        y_pred = y_pred.astype(np.bool)
        if category_dim is None:
            return np.mean(x[y_pred]) - np.mean(x[y_true])

        assert category_dim == -1
        num_classes = y_pred.shape[-1]

        hu_pred = np.array([np.mean(x[y_pred[..., c]]) for c in range(num_classes)])
        hu_true = np.array([np.mean(x[y_true[..., c]]) for c in range(num_classes)])

        vals = hu_pred - hu_true
        if self._method == "percent":
            vals = vals / np.absolute(hu_true) * 100

        return (vals,)

    def name(self):
        symbol = "Abs" if self._method == "absolute" else "%"
        return "HU {}Diff".format(symbol)


class AreaDiff(Metric):
    """
    Calculates the absolute difference or the percent error in cross-sectional area
    when compared to the ground truth.
    """

    def __init__(self, units="", method="absolute"):
        super().__init__(units)
        assert method in ["absolute", "percent"]
        self._method = method

    def __call__(self, y_pred, y_true, spacing=None, category_dim: int = None):
        pixel_area = np.prod(spacing) if spacing else 1
        y_pred = y_pred.astype(np.bool)
        y_true = y_true.astype(np.bool)
        y_pred, y_true = flatten_non_category_dims((y_pred, y_true), category_dim)

        size_pred = pixel_area * np.count_nonzero(y_pred, -1)
        size_true = pixel_area * np.count_nonzero(y_true, -1)

        val = size_pred - size_true
        if self._method == "percent":
            val = 100 * val / size_true

        return (val,)

    def name(self):
        symbol = "Abs" if self._method == "absolute" else "%"
        return "Area {}Diff".format(symbol)


@EVALUATOR_REGISTRY.register()
class CTEvaluator(SemSegEvaluator):
    """Evaluate semantic segmentation on CT datasets.

    Includes Hounsfield unit difference and pixel area difference between datasets.
    """

    def _get_metrics(self):
        metrics = list(super()._get_metrics())
        paired_metrics = [
            HU(method="pred"),
            HU(method="base"),
            CSA(method="pred", units="(mm^2)"),
            CSA(method="base", units="(mm^2)"),
        ]
        metrics.extend(
            [
                HUDiff(method="absolute"),
                HUDiff(method="percent"),
                AreaDiff(method="absolute", units="(mm^2)"),
                AreaDiff(method="percent"),
            ]
        )
        metrics.extend(paired_metrics)

        self._metric_pairs = [
            (p.name(), b.name()) for p, b in zip(paired_metrics[::2], paired_metrics[1::2])
        ]
        return metrics

    def reset(self):
        super().reset()
        for pred_key, base_key in self._metric_pairs:
            self._metrics_manager.register_pairs(
                pred_key, base_key, (Reductions.RMS_CV, Reductions.RMSE_CV)
            )

    def process(self, inputs, outputs):
        if not self.spacing:
            assert all("scan_spacing" in x for x in inputs)
        super().process(inputs, outputs)


@EVALUATOR_REGISTRY.register()
class CTEvaluatorTTPP(CTEvaluator):
    """CT Evaluator with test-time post-processing."""

    def __init__(
        self,
        dataset_name: str,
        cfg,
        output_folder: str = None,
        save_raw_data: bool = False,
        stream_evaluation: bool = True,
    ):
        super().__init__(
            dataset_name,
            cfg,
            os.path.join(cfg.OUTPUT_DIR, "test_results_pp-muscle-imat"),
            save_raw_data,
            stream_evaluation,
        )

    def _postprocess_labels(self, x, labels):
        labels = labels.copy()
        categories = self._categories
        muscle_idx = categories.index("muscle")
        imat_idx = categories.index("intramuscular fat")

        muscle_mask = labels[..., muscle_idx]
        imat_mask = labels[..., imat_idx]

        imat_mask[muscle_mask.astype(np.bool) & (x < -30)] = 1
        muscle_mask[x < -30] = 0

        labels[..., muscle_idx] = muscle_mask
        labels[..., imat_idx] = imat_mask

        return labels

    def process(self, inputs, outputs):
        if not self.spacing:
            assert all("scan_spacing" in x for x in inputs)

        output_activation = self._output_activation
        includes_bg = self._output_includes_background

        for input, output in zip(inputs, outputs):
            y_pred = output["y_pred"]

            if output_activation == "sigmoid":
                labels = (y_pred > 0.5).astype(np.uint8)
            elif output_activation == "softmax":
                labels = np.zeros_like(y_pred, dtype=np.uint8)
                l_argmax = np.argmax(y_pred, axis=-1)
                for c in range(labels.shape[-1]):
                    labels[l_argmax == c, c] = 1
                labels = labels.astype(np.uint)
            else:
                raise ValueError("output activation {} not supported".format(output_activation))

            # background is always excluded from analysis
            if includes_bg:
                y_true = output["y_true"][..., 1:]
                y_pred = output["y_pred"][..., 1:]
                labels = labels[..., 1:]
                output["y_true"] = y_true
                output["y_pred"] = y_pred

            # Post process labels to switch muscle/imat labels at hu threshold of -40
            labels = self._postprocess_labels(input["x"], labels)

            time_elapsed = output["time_elapsed"]
            if self.stream_evaluation:
                self.eval_single_scan(input, output, labels, time_elapsed)
            else:
                self._predictions.append((input, output, labels, time_elapsed))



import unittest

from keras import backend as K
from keras.layers import Activation, Concatenate, Conv3D, Conv3DTranspose

from medsegpy.config import UNet2_5DConfig, UNet3DConfig, UNetConfig
from medsegpy.modeling.meta_arch import build
from medsegpy.utils import env

if not env.is_tf2():
    _TF2 = False
    from medsegpy.modeling.build import get_model
else:
    _TF2 = True
    get_model = None


class TestUNet2D(unittest.TestCase):
    """Test building 2D/2.5D UNet using the builder."""

    @classmethod
    def _compare_layers(cls, m1, m2):
        for l1, l2 in zip(m1.layers, m2.layers):
            assert type(l1) == type(l2)  # noqa

            l1_cfg = l1.get_config()
            l2_cfg = l2.get_config()
            l2_cfg.pop("name"), l1_cfg.pop("name")

            if isinstance(l2, Concatenate):
                channel_axis = -1 if K.image_data_format() == "channels_last" else 1  # noqa
                mapping = {channel_axis: 3}
                l2_cfg["axis"] = mapping[l2_cfg["axis"]]

            assert l1_cfg == l2_cfg, "{}: {}\n{}: {}".format(
                l1.name, l1.get_config(), l2.name, l2.get_config()
            )

    def test_same_unet2d(self):
        """2D U-Net should be same between builder and function construct."""
        if _TF2:
            return
        cfg = UNetConfig()
        cfg.CATEGORIES = [0]
        m1 = get_model(cfg)
        m2 = build.build_model(cfg)

        self._compare_layers(m1, m2)

    def test_same_unet_2_5d(self):
        """2.5D U-Net should be same between builder and function construct.

        The builder takes in the 2D U-Net config and constructs a 2.5 config.
        This motivates deprecating UNet2_5DConfig.
        """
        if _TF2:
            return
        cfg1 = UNet2_5DConfig()
        cfg1.IMG_SIZE = (288, 288, 3)
        cfg1.CATEGORIES = [0]
        cfg2 = UNetConfig()
        cfg2.IMG_SIZE = (288, 288, 3)
        cfg2.CATEGORIES = [0]
        m1 = get_model(cfg1)
        m2 = build.build_model(cfg2)

        self._compare_layers(m1, m2)


class TestUNet3D(unittest.TestCase):
    """Test building 3D UNet using the builder."""

    def test_same_unet_3d(self):
        """3D U-Net should be same between builder and function construct."""
        if _TF2:
            return
        cfg = UNet3DConfig()
        cfg.CATEGORIES = [0]
        m1 = get_model(cfg)
        m2 = build.build_model(cfg)

        l1_idx, l2_idx = 0, 0

        while True:
            reached_l1 = l1_idx == len(m1.layers)
            reached_l2 = l2_idx == len(m2.layers)
            if reached_l1 and reached_l2:
                break

            assert not reached_l1 and not reached_l2, "l1_idx: {}, l2_idx: {}".format(
                l1_idx, l2_idx
            )

            l1 = m1.layers[l1_idx]
            l1_cfg = l1.get_config()
            l2 = m2.layers[l2_idx]
            l2_cfg = l2.get_config()

            l1_cfg.pop("name"), l2_cfg.pop("name")

            if isinstance(l2, Concatenate):
                channel_axis = -1 if K.image_data_format() == "channels_last" else 1  # noqa
                mapping = {channel_axis: 4}
                l2_cfg["axis"] = mapping[l2_cfg["axis"]]
            elif isinstance(l1, Conv3D) and not isinstance(l1, Conv3DTranspose):
                # Handle conv+activation case.
                l1_cfg.pop("activation")
                l1_activation = m1.layers[l1_idx + 1]
                assert isinstance(
                    l1_activation, Activation
                ), "Expected Activation - l1_idx: {}: {}\n{}".format(
                    l1_idx, type(l1_activation), m1.layers[l1_idx - 1 : l1_idx + 1]
                )
                l1_activation = l1_activation.get_config()["activation"]
                l2_activation = l2_cfg.pop("activation")
                assert l1_activation == l2_activation, "l1_idx: {} - {}\nl2_cfg: {} - {}".format(
                    l1_idx, l1_cfg, l2_idx, l2_cfg
                )
                l1_idx += 1

            assert l1_cfg == l2_cfg, "{}: {}\n{}: {}".format(l1.name, l1_cfg, l2.name, l2_cfg)

            l1_idx += 1
            l2_idx += 1


if __name__ == "__main__":
    unittest.main()
